"""Utility functions for SageMaker model monitoring"""

import json
import logging
import pandas as pd
from typing import Dict, Any, Union
from io import StringIO
import awswrangler as wr

# Workbench Imports
from workbench.utils.s3_utils import read_content_from_s3

# Setup logging
log = logging.getLogger("workbench")


def pull_data_capture(data_capture_path, max_files=1) -> Union[pd.DataFrame, None]:
    """
    Read and process captured data from S3.

    Args:
        data_capture_path (str): S3 path to the data capture files.
        max_files (int, optional): Maximum number of files to process.
                                  Defaults to 1 (most recent only).
                                  Set to None to process all files.

    Returns:
        Union[pd.DataFrame, None]: A dataframe of the captured data (or None if no data is found).
    """
    # List files in the specified S3 path
    files = wr.s3.list_objects(data_capture_path)
    if not files:
        log.warning(f"No data capture files found in {data_capture_path}.")
        return None

    log.info(f"Found {len(files)} files in {data_capture_path}.")

    # Sort files by timestamp (assuming the naming convention includes timestamp)
    files.sort()

    # Select files to process
    if max_files is None:
        files_to_process = files
        log.info(f"Processing all {len(files)} files.")
    else:
        files_to_process = files[-max_files:] if files else []
        log.info(f"Processing the {len(files_to_process)} most recent file(s).")

    # Process each file
    all_data = []
    for file_path in files_to_process:
        try:
            # Read the JSON lines file
            df = wr.s3.read_json(path=file_path, lines=True)
            if not df.empty:
                all_data.append(df)
        except Exception as e:
            log.warning(f"Error processing file {file_path}: {e}")

    # Combine all DataFrames and return
    return pd.concat(all_data, ignore_index=True)


def process_data_capture(df: pd.DataFrame) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Process the captured data DataFrame to extract input and output data.
    Continues processing even if individual files are malformed.
    Args:
        df (DataFrame): DataFrame with captured data.
    Returns:
        tuple[DataFrame, DataFrame]: Input and output DataFrames.
    """
    input_dfs = []
    output_dfs = []

    for idx, row in df.iterrows():
        try:
            capture_data = row["captureData"]

            # Check if this capture has the required fields (all or nothing)
            if "endpointInput" not in capture_data:
                log.warning(f"Row {idx}: No endpointInput found in capture data.")
                continue

            if "endpointOutput" not in capture_data:
                log.critical(
                    f"Row {idx}: No endpointOutput found in capture data. DataCapture needs to include Output capture!"
                )
                continue

            # Process input data
            input_data = capture_data["endpointInput"]
            if input_data["encoding"].upper() == "CSV":
                input_df = pd.read_csv(StringIO(input_data["data"]))
            elif input_data["encoding"].upper() == "JSON":
                json_data = json.loads(input_data["data"])
                if isinstance(json_data, dict):
                    input_df = pd.DataFrame({k: [v] if not isinstance(v, list) else v for k, v in json_data.items()})
                else:
                    input_df = pd.DataFrame(json_data)

            # Process output data
            output_data = capture_data["endpointOutput"]
            if output_data["encoding"].upper() == "CSV":
                output_df = pd.read_csv(StringIO(output_data["data"]))
            elif output_data["encoding"].upper() == "JSON":
                json_data = json.loads(output_data["data"])
                if isinstance(json_data, dict):
                    output_df = pd.DataFrame({k: [v] if not isinstance(v, list) else v for k, v in json_data.items()})
                else:
                    output_df = pd.DataFrame(json_data)

            # If we get here, both processed successfully
            input_dfs.append(input_df)
            output_dfs.append(output_df)

        except Exception as e:
            log.error(f"Row {idx}: Failed to process row: {e}")
            continue
    # Combine and return results
    return (
        pd.concat(input_dfs, ignore_index=True) if input_dfs else pd.DataFrame(),
        pd.concat(output_dfs, ignore_index=True) if output_dfs else pd.DataFrame(),
    )


def get_monitor_json_data(s3_path: str) -> Union[dict, None]:
    """
    Convert JSON monitoring data into a DataFrame

    Args:
        s3_path (str): The S3 path to the monitoring data

    Returns:
        dict: A dictionary of the monitoring data (None if not found)
    """
    # Check if the S3 path exists
    if not wr.s3.does_object_exist(path=s3_path):
        log.warning(f"Monitoring data does not exist in S3: {s3_path}")
        return None

    # Read the JSON data from S3
    raw_json = read_content_from_s3(s3_path)
    return json.loads(raw_json)


def parse_monitoring_results(results_json: str) -> Dict[str, Any]:
    """
    Parse monitoring results from JSON

    Args:
        results_json (str): Monitoring results in JSON format

    Returns:
        dict: Parsed monitoring results
    """
    try:
        results = json.loads(results_json)

        # Extract and format the key information
        parsed_results = {
            "schema_validation": results.get("schema", {}).get("validation", {}),
            "constraint_violations": [],
        }

        # Extract violations
        for violation in results.get("violations", []):
            parsed_violation = {
                "feature_name": violation.get("feature_name"),
                "constraint_check_type": violation.get("constraint_check_type"),
                "description": violation.get("description"),
            }
            parsed_results["constraint_violations"].append(parsed_violation)

        return parsed_results
    except Exception as e:
        log.error(f"Error parsing monitoring results: {e}")
        return {"error": str(e)}


"""TEMP
                # If the status is "CompletedWithViolations", we grab the lastest
                # violation file and add it to the result
                if status == "CompletedWithViolations":
                    violation_file = f"{self.monitoring_path}/
                    {last_run['CreationTime'].strftime('%Y/%m/%d')}/constraint_violations.json"
                    if wr.s3.does_object_exist(violation_file):
                        violations_json = read_content_from_s3(violation_file)
                        violations = parse_monitoring_results(violations_json)
                        result["violations"] = violations.get("constraint_violations", [])
                        result["violation_count"] = len(result["violations"])
                    else:
                        result["violations"] = []
                        result["violation_count"] = 0
"""


def preprocessing_script(feature_list: list[str]) -> str:
    """
    A preprocessing script for monitoring jobs.

    Args:
        feature_list (list[str]): List of features to include in the preprocessing script.

    Returns:
        str: The preprocessing script
    """
    # Convert feature list to a proper Python string representation
    features_str = str(feature_list)

    script = """
import pandas as pd
from io import StringIO

def preprocess_handler(inference_record, logger):

    # CapturedData objects have endpoint_input with encoding and data attributes
    input_data = inference_record.endpoint_input.data

    # Parse the input data (assuming CSV format)
    df = pd.read_csv(StringIO(input_data))
    logger.info("Input DataFrame:")
    logger.info(df.shape)
    logger.info(df.columns)

    # Keep only the specified features
    # Note: That this feature list needs to be alphabetically sorted
    feature_list = {{features_str}}
    df = df[feature_list]

    logger.info("Output DataFrame:")
    logger.info(df.shape)
    logger.info(df.columns)
    output_data = df.to_dict(orient='records')
    return output_data
"""
    # Replace the placeholder with the actual feature list
    script = script.replace("{{features_str}}", features_str)
    return script


# Test function for the utils
if __name__ == "__main__":
    """Test the monitor_utils module"""
    from workbench.api.monitor import Monitor

    # Test pulling data capture
    mon = Monitor("caco2-pappab-class-0")
    df = pull_data_capture(mon.data_capture_path)
    print("Data Capture:")
    print(df.head())

    # Test processing data capture
    input_processed, output_processed = process_data_capture(df)

    print("\nProcessed Input:")
    print(input_processed)

    print("\nProcessed Output:")
    print(output_processed)

    # Test preprocessing script
    script = preprocessing_script(["feature1", "feature2", "feature3"])
    print("\nPreprocessing Script:")
    print(script)



import numpy as np
import matplotlib.pyplot as plt
import statsmodels.api as sm

# Example data
np.random.seed(42)
x = np.linspace(0, 10, 100)
y_true = 3 * x + np.random.normal(0, 1, 100)
y_pred = 3 * x

# Residuals
residuals = y_true - y_pred

# Plot Residuals vs. Fitted Values
plt.figure(figsize=(10, 6))
plt.scatter(y_pred, residuals)
plt.axhline(0, color="red", linestyle="--")
plt.xlabel("Fitted Values")
plt.ylabel("Residuals")
plt.title("Residuals vs. Fitted Values")
plt.show()

# Q-Q Plot
sm.qqplot(residuals, line="45")
plt.title("Q-Q Plot")
plt.show()



"""This Script creates an empty model group, a model group without any models

Models:
    - empty-model-group
"""

import logging
from botocore.exceptions import ClientError
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp

# Setup the logger
log = logging.getLogger("workbench")


# Helper method to create an 'empty' model group
def create_model_package_group(group_name, boto3_session):
    sm_client = boto3_session.client("sagemaker")

    try:
        response = sm_client.create_model_package_group(
            ModelPackageGroupName=group_name,
            ModelPackageGroupDescription="Your description here",  # Add a relevant description
        )
        log.info(f"Model Package Group '{group_name}' created successfully.")
        return response
    except ClientError as e:
        if e.response["Error"]["Code"] == "ValidationException" and "already exists" in e.response["Error"]["Message"]:
            log.info(f"Model Package Group '{group_name}' already exists.")
            return None
        else:
            # Re-raise any other exceptions
            raise


if __name__ == "__main__":

    # Create an empty Model Package Group
    my_boto3_session = AWSAccountClamp().boto3_session
    create_model_package_group("empty-model-group", my_boto3_session)



"""Tests for confusion_matrix web component"""

# Workbench Imports
from workbench.web_interface.components.plugins.confusion_matrix import ConfusionMatrix
from workbench.api.model import Model


def test_confusion_matrix():
    """Test the ConfusionMatrix class"""

    # Instantiate model
    model = Model("wine-classification")
    inference_run = "auto_inference"

    # Instantiate the ConfusionMatrix class
    cm = ConfusionMatrix()

    # Update the properties (the figure is the first element in the property list)
    all_properties = cm.update_properties(model, inference_run=inference_run)
    figure = all_properties[0]
    figure.show()


if __name__ == "__main__":
    # Run the tests
    test_confusion_matrix()



"""AWSGraphStore: Storage of NetworkX Graphs using AWS S3 and JSON serialization."""

import networkx as nx
import json
import logging
import re
from datetime import datetime
from typing import Union, Optional
from urllib.parse import urlparse
import pandas as pd

# Workbench Imports
from workbench.utils.config_manager import ConfigManager
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp
from workbench.utils.json_utils import CustomEncoder, custom_decoder


class AWSGraphStore:
    """AWSGraphStore: Storage of NetworkX Graphs using AWS S3 and JSON serialization."""

    def __init__(self, path_prefix: Optional[str] = None):
        """AWSGraphStore Init Method

        Args:
            path_prefix (Optional[str]): Path prefix for storage locations (Defaults to None)
        """
        self.log = logging.getLogger("workbench")
        self._base_prefix = "graph_store/"
        self.path_prefix = self._normalize_path(self._base_prefix + (path_prefix or ""))

        # Retrieve bucket name and initialize S3 session
        config = ConfigManager()
        self.workbench_bucket = config.get_config("WORKBENCH_BUCKET")
        self.boto3_session = AWSAccountClamp().boto3_session
        self.s3_client = self.boto3_session.client("s3")

    def summary(self) -> pd.DataFrame:
        """Provide a summary of all graphs in the store.

        Returns:
            pd.DataFrame: Summary DataFrame with location, size, and modified date.
        """
        df = self.details()
        if df.empty:
            return pd.DataFrame(columns=["location", "size (MB)", "modified"])

        df["size (MB)"] = (df["size"] / (1024 * 1024)).round(2)
        df["modified"] = pd.to_datetime(df["modified"]).dt.strftime("%Y-%m-%d %H:%M:%S")
        return df[["location", "size (MB)", "modified"]]

    def details(self) -> pd.DataFrame:
        """Return detailed metadata for all stored graphs.

        Returns:
            pd.DataFrame: DataFrame with details like location, size, and last modified date.
        """
        try:
            response = self.s3_client.list_objects_v2(Bucket=self.workbench_bucket, Prefix=self.path_prefix)
            if "Contents" not in response:
                return pd.DataFrame(columns=["location", "s3_file", "size", "modified"])

            data = [
                {
                    "location": obj["Key"].replace(f"{self.path_prefix}", "").split(".json")[0],
                    "s3_file": f"s3://{self.workbench_bucket}/{obj['Key']}",
                    "size": obj["Size"],
                    "modified": obj["LastModified"],
                }
                for obj in response["Contents"]
            ]
            return pd.DataFrame(data)
        except Exception as e:
            self.log.error(f"Failed to get object details: {e}")
            return pd.DataFrame(columns=["location", "s3_file", "size", "modified"])

    def check(self, location: str) -> bool:
        """Check if a graph exists.

        Args:
            location (str): Logical location of the graph.

        Returns:
            bool: True if the graph exists, False otherwise.
        """
        s3_uri = self._generate_s3_uri(location)
        bucket, key = self._parse_s3_uri(s3_uri)

        try:
            self.s3_client.head_object(Bucket=bucket, Key=key)
            return True
        except self.s3_client.exceptions.ClientError:
            return False

    def get(self, location: str) -> Union[nx.Graph, None]:
        """Retrieve a NetworkX graph from AWS S3.

        Args:
            location (str): Logical location of the graph.

        Returns:
            Union[nx.Graph, None]: The retrieved graph or None if not found.
        """
        s3_uri = self._generate_s3_uri(location)
        bucket, key = self._parse_s3_uri(s3_uri)

        try:
            response = self.s3_client.get_object(Bucket=bucket, Key=key)
            graph_json = json.loads(response["Body"].read().decode("utf-8"), object_hook=custom_decoder)

            # Deserialize the graph
            graph = nx.readwrite.json_graph.node_link_graph(graph_json, edges="edges")

            # Replace "_node_id" back to "id" in the graph's node attributes
            for node in graph.nodes:
                if "_node_id" in graph.nodes[node]:
                    graph.nodes[node]["id"] = graph.nodes[node].pop("_node_id")
            return graph
        except json.JSONDecodeError as e:
            self.log.error(f"Failed to decode JSON for graph at '{s3_uri}': {e}")
            raise
        except Exception as e:
            self.log.error(f"Failed to retrieve graph from '{s3_uri}': {e}")
            return None

    def upsert(self, location: str, graph: nx.Graph):
        """Insert or update a NetworkX graph in AWS S3.

        Args:
            location (str): Logical location to store the graph.
            graph (nx.Graph): The NetworkX graph to store.
        """
        s3_uri = self._generate_s3_uri(location)
        bucket, key = self._parse_s3_uri(s3_uri)

        try:
            graph_json = nx.readwrite.json_graph.node_link_data(graph, edges="edges")

            # If we have an "id" field, replicate that into "_node_id" for serialization
            for node in graph_json["nodes"]:
                if "id" in node:
                    node["_node_id"] = node["id"]

            json_data = json.dumps(graph_json, cls=CustomEncoder)
            self.s3_client.put_object(Bucket=bucket, Key=key, Body=json_data)
            self.log.info(f"Graph stored at '{s3_uri}'")
        except Exception as e:
            self.log.error(f"Failed to store graph at '{s3_uri}': {e}")
            raise

    def list(self) -> list:
        """List all graphs in the store.

        Returns:
            list: A list of all graph locations in the store.
        """
        df = self.details()
        return df["location"].tolist()

    def last_modified(self, location: str) -> Union[datetime, None]:
        """Return the last modified date of a graph.

        Args:
            location (str): Logical location of the graph.

        Returns:
            Union[datetime, None]: Last modified datetime or None if not found.
        """
        s3_uri = self._generate_s3_uri(location)
        bucket, key = self._parse_s3_uri(s3_uri)

        try:
            response = self.s3_client.head_object(Bucket=bucket, Key=key)
            return response["LastModified"]
        except self.s3_client.exceptions.ClientError:
            return None

    def delete(self, location: str):
        """Delete a NetworkX graph from AWS S3.

        Args:
            location (str): Logical location of the graph to delete.
        """
        s3_uri = self._generate_s3_uri(location)
        bucket, key = self._parse_s3_uri(s3_uri)

        try:
            self.s3_client.delete_object(Bucket=bucket, Key=key)
            self.log.info(f"Graph deleted from '{s3_uri}'")
        except Exception as e:
            self.log.error(f"Failed to delete graph from '{s3_uri}': {e}")

    def _generate_s3_uri(self, location: str) -> str:
        """Generate the S3 URI for the given location."""
        s3_path = f"{self.workbench_bucket}/{self.path_prefix}/{location}.json"
        return f"s3://{re.sub(r'/+', '/', s3_path)}"

    def _parse_s3_uri(self, s3_uri: str) -> tuple:
        """Parse an S3 URI into bucket and key."""
        parsed = urlparse(s3_uri)
        if parsed.scheme != "s3":
            raise ValueError(f"Invalid S3 URI: {s3_uri}")
        return parsed.netloc, parsed.path.lstrip("/")

    @staticmethod
    def _normalize_path(path: str) -> str:
        """Normalize paths by collapsing slashes."""
        return re.sub(r"/+", "/", path)

    def __repr__(self):
        """Return a string representation of the AWSGraphStore object."""
        summary_df = self.summary()
        if summary_df.empty:
            return "AWSGraphStore: No graphs found in the store."
        return summary_df.to_string(index=False)


if __name__ == "__main__":
    """Exercise the AWSGraphStore Class"""
    # Create an AWSGraphStore instance
    graph_store = AWSGraphStore()

    # Create a test graph
    G = nx.erdos_renyi_graph(5, 0.5)
    for u, v, d in G.edges(data=True):
        d["weight"] = 1.0

    # Add node attributes, including an "id"
    for node in G.nodes:
        G.nodes[node]["id"] = node

    # Store the graph
    graph_store.upsert("test/test_graph", G)

    # Get the graph and print out attributes
    G = graph_store.get("test/test_graph")
    print("Graph Nodes:", G.nodes(data=True))

    # Get a summary of the graphs
    print("Graph Store Summary:")
    print(graph_store.summary())

    # Display detailed metadata
    print("Graph Store Details:")
    print(graph_store.details())

    # Last modified date
    print("Last modified:", graph_store.last_modified("test/test_graph"))

    # List all graphs
    print("List all graphs in the store:")
    print(graph_store.list())

    # Retrieve and display the graph
    retrieved_graph = graph_store.get("test/test_graph")
    print("Retrieved Graph Edges:", retrieved_graph.edges(data=True))

    # Check if the graph exists
    print("Graph exists:", graph_store.check("test/test_graph"))

    # Delete the graph
    graph_store.delete("test/test_graph")
    print("Graph exists after deletion:", graph_store.check("test/test_graph"))



"""Callbacks for the Pipelines Subpage Dashboard Interface"""

import logging

from dash import callback, Output, Input, State
from dash.exceptions import PreventUpdate
from urllib.parse import urlparse, parse_qs


# Workbench Imports
from workbench.web_interface.page_views.pipelines_page_view import PipelinesPageView
from workbench.web_interface.components.plugins.ag_table import AGTable
from workbench.cached.cached_pipeline import CachedPipeline

# Get the Workbench logger
log = logging.getLogger("workbench")


def on_page_load():
    @callback(
        Output("pipelines_table", "selectedRows"),
        Output("pipelines_page_loaded", "data"),
        Input("url", "href"),
        Input("pipelines_table", "rowData"),
        State("pipelines_page_loaded", "data"),
        prevent_initial_call=True,
    )
    def _on_page_load(href, row_data, page_already_loaded):
        if page_already_loaded:
            raise PreventUpdate

        if not href or not row_data:
            raise PreventUpdate

        parsed = urlparse(href)
        if parsed.path != "/pipelines":
            raise PreventUpdate

        selected_name = parse_qs(parsed.query).get("name", [None])[0]
        if not selected_name:
            return [row_data[0]], True

        for row in row_data:
            if row.get("name") == selected_name:
                return [row], True

        raise PreventUpdate


def pipeline_table_refresh(page_view: PipelinesPageView, table: AGTable):
    @callback(
        [Output(component_id, prop) for component_id, prop in table.properties],
        Input("pipelines_refresh", "n_intervals"),
    )
    def _pipeline_table_refresh(_n):
        """Return the table data for the Pipelines Table"""
        page_view.refresh()
        pipelines = page_view.pipelines()
        pipelines["name"] = pipelines["Name"]
        pipelines["id"] = range(len(pipelines))
        return table.update_properties(pipelines)


# Set up the plugin callbacks that take a pipeline
def setup_plugin_callbacks(plugins):
    @callback(
        # Aggregate plugin outputs
        [Output(component_id, prop) for p in plugins for component_id, prop in p.properties],
        Input("pipelines_table", "selectedRows"),
    )
    def update_all_plugin_properties(selected_rows):
        # Check for no selected rows
        if not selected_rows:
            raise PreventUpdate

        # Get the selected row data and grab the name
        selected_row_data = selected_rows[0]
        pipeline_name = selected_row_data["Name"]

        # Create the Endpoint object
        pipeline = CachedPipeline(pipeline_name)

        # Update all the properties for each plugin
        all_props = []
        for p in plugins:
            all_props.extend(p.update_properties(pipeline))

        # Return all the updated properties
        return all_props



import boto3
import aws_cdk as cdk
from aws_cdk import Fn
from workbench_compute.workbench_compute_stack import WorkbenchComputeStack, WorkbenchComputeStackProps

# Initialize Workbench ConfigManager
try:
    from workbench.utils.config_manager import ConfigManager

    cm = ConfigManager()
except ImportError:
    print("Workbench Configuration Manager Not Found...")
    print("Set the WORKBENCH_CONFIG Env var and run again...")
    raise SystemExit(1)

# Grab the account and region using boto3
session = boto3.session.Session()
aws_account = session.client("sts").get_caller_identity().get("Account")
aws_region = session.region_name
print(f"Account: {aws_account}")
print(f"Region: {aws_region}")

# Get configurations needed for compute stack
workbench_bucket = cm.get_config("WORKBENCH_BUCKET")
if not workbench_bucket:
    print("Error: WORKBENCH_BUCKET is required but not found in config.")
    exit(1)


# Get the Batch and Lambda Roles from WorkbenchCore stack outputs
batch_role_arn = Fn.import_value("WorkbenchCore-BatchRoleArn")
lambda_role_arn = Fn.import_value("WorkbenchCore-LambdaRoleArn")

# VPC and subnet configuration (optional)
existing_vpc_id = cm.get_config("WORKBENCH_VPC_ID")
subnet_ids = cm.get_config("WORKBENCH_SUBNET_IDS") or []

# Log the configuration for transparency
print("Configuration:")
print(f"  WORKBENCH_BUCKET: {workbench_bucket}")
print(f"  WORKBENCH_BATCH_ROLE_ARN: {batch_role_arn}")
print(f"  WORKBENCH_LAMBDA_ROLE_ARN: {lambda_role_arn}")
print(f"  WORKBENCH_VPC_ID: {existing_vpc_id}")
print(f"  WORKBENCH_SUBNET_IDS: {subnet_ids}")

# Our CDK App and Environment
app = cdk.App()
env = cdk.Environment(account=aws_account, region=aws_region)

# Create the Workbench Compute Stack
compute_stack = WorkbenchComputeStack(
    app,
    "WorkbenchCompute",
    env=env,
    props=WorkbenchComputeStackProps(
        workbench_bucket=workbench_bucket,
        batch_role_arn=batch_role_arn,
        lambda_role_arn=lambda_role_arn,
        existing_vpc_id=existing_vpc_id,
        subnet_ids=subnet_ids,
    ),
)

app.synth()



# Workbench Imports
from workbench.api.df_store import DFStore
from workbench.utils.chem_utils import compute_morgan_fingerprints
from workbench.algorithms.dataframe.fingerprint_proximity import FingerprintProximity

# Grab tox21 dataset (only 10)
tox_df = DFStore().get("/datasets/chem_info/tox21_10")
# tox_df = tox_df.sample(100)

# Compute Morgan fingerprints
tox_df = compute_morgan_fingerprints(tox_df)

# Compute FingerprintProximity
prox = FingerprintProximity(tox_df, fingerprint_column="fingerprint", id_column="id", n_neighbors=5)

# Get all the neighbors for all the compounds
neighbors_df = prox.all_neighbors()
print("\nAll neighbors:")
print(neighbors_df.head(20))

# Query for neighbors for a specific compound
query_id = tox_df["id"].iloc[1]
query_neighbors_df = prox.neighbors(query_id=query_id)
print(f"\nKNN: Neighbors for query ID {query_id}:")
print(query_neighbors_df)



import requests


def get_pubchem_summary_dynamic(cid: int) -> dict:
    """Fetch summary information for a compound from PubChem."""
    url = f"https://pubchem.ncbi.nlm.nih.gov/rest/pug_view/data/compound/{cid}/JSON"
    response = requests.get(url)
    response.raise_for_status()
    data = response.json()

    summary = {"Name": None, "Molecular Formula": None, "Molecular Weight": None, "Toxicity": []}

    # Traverse the JSON dynamically to find relevant fields
    sections = data.get("Record", {}).get("Section", [])
    for section in sections:
        if section.get("TOCHeading") == "Names and Identifiers":
            for sub_section in section.get("Section", []):
                if sub_section.get("TOCHeading") == "Record Title":
                    summary["Name"] = (
                        sub_section.get("Information", [{}])[0]
                        .get("Value", {})
                        .get("StringWithMarkup", [{}])[0]
                        .get("String")
                    )

        if section.get("TOCHeading") == "Chemical and Physical Properties":
            for sub_section in section.get("Section", []):
                if sub_section.get("TOCHeading") == "Molecular Formula":
                    summary["Molecular Formula"] = (
                        sub_section.get("Information", [{}])[0]
                        .get("Value", {})
                        .get("StringWithMarkup", [{}])[0]
                        .get("String")
                    )
                if sub_section.get("TOCHeading") == "Molecular Weight":
                    summary["Molecular Weight"] = (
                        sub_section.get("Information", [{}])[0]
                        .get("Value", {})
                        .get("StringWithMarkup", [{}])[0]
                        .get("String")
                    )

        if section.get("TOCHeading") == "Toxicity":
            for sub_section in section.get("Section", []):
                if "Toxicological Information" in sub_section.get("TOCHeading", ""):
                    for info in sub_section.get("Information", []):
                        toxicity_info = info.get("Value", {}).get("StringWithMarkup", [{}])[0].get("String")
                        summary["Toxicity"].append(toxicity_info)

    # Return the cleaned summary with any found values
    return {k: v for k, v in summary.items() if v}


# Example usage
cid = 240  # Ethanol
summary_info = get_pubchem_summary_dynamic(cid)
print(summary_info)



import pytest
from workbench.api.data_source import DataSource
from pprint import pprint


@pytest.mark.long
def test():
    # Create a new Data Source from an S3 Path (or a local file)
    source_path = "s3://workbench-public-data/common/abalone.csv"
    my_data = DataSource(source_path)
    pprint(my_data.summary())
    pprint(my_data.details())

    # Create a FeatureSet (with a name that has mixed case)
    fs = my_data.to_features("aBaLone-feaTures-Temp")
    assert fs is None


if __name__ == "__main__":
    test()



"""FeatureResolution: Report on Feature Space Resolution Issues"""

import logging
from typing import Union
import pandas as pd
from sklearn.neighbors import KNeighborsRegressor
from sklearn.preprocessing import StandardScaler

# Workbench Imports
from workbench.utils.pandas_utils import DataFrameBuilder


# Feature Resolution Class
class FeatureResolution:
    def __init__(
        self,
        input_df: pd.DataFrame,
        features: list,
        target_column: str,
        id_column: str,
        distance_metric: str = "minkowski",
    ):
        """Initialize the FeatureResolution object

        Args:
            distance_metric: Distance metric to use (default: "minkowski")
        """
        self.log = logging.getLogger("workbench")
        self.df = input_df.copy().reset_index(drop=True)
        self.features = features
        self.target_column = target_column
        self.id_column = id_column
        self.n_neighbors = 10
        self.scalar = StandardScaler()
        self.knn = KNeighborsRegressor(metric=distance_metric, n_neighbors=self.n_neighbors, weights="distance")
        self.dataframe_builder = DataFrameBuilder()
        self.recursive_df_list = []

    def compute(
        self, within_distance: float, min_target_difference: float, output_columns: list = [], verbose=True
    ) -> Union[pd.DataFrame, None]:
        """FeatureResolution: Compute Feature Space to Target Resolution and Report Issues

        Args:
            within_distance: Features within this distance should have similar target values
            min_target_difference: Minimum target difference to consider
            output_columns: List of additional columns to output (default: []])
            verbose: Whether to print out the resolution issues (default: True)

        Returns:
            Pandas DataFrame of Feature Space to Target Resolution Issues
            Includes any additional output_columns if specified
        """

        # Check for expected columns
        for column in [self.target_column] + self.features:
            if column not in self.df.columns:
                self.log.error(f"DataFrame does not have required {column} Column!")
                return

        # Set up the output columns (add id and target columns if they are not already included)
        output_columns = list(set(output_columns).union({self.id_column, self.target_column}))

        # Check the output columns
        if output_columns is not None:
            for column in output_columns:
                if column not in self.df.columns:
                    self.log.error(f"DataFrame does not have required {column} Column!")
                    return

        # Check for NaNs in the features and log the percentage
        for feature in self.features:
            nan_count = self.df[feature].isna().sum()
            if nan_count > 0:
                print(f"Feature '{feature}' has {nan_count} NaNs ({nan_count / len(self.df) * 100:.2f}%).")

        # Remove and NaNs or INFs in the features
        self.log.info(f"Dataframe Shape before NaN/INF removal {self.df.shape}")
        self.df = self.df.replace([float("inf"), float("-inf")], pd.NA).dropna().reset_index(drop=True)
        self.log.info(f"Dataframe Shape after NaN/INF removal {self.df.shape}")

        # Standardize the features
        X = self.scalar.fit_transform(self.df[self.features])
        y = self.df[self.target_column]

        # Fit the KNN model
        self.knn.fit(X, y)

        # Compute the feature space to target resolution to the nearest neighbors
        output_count = 0
        for my_index, row in enumerate(X):
            # Find the nearest neighbors
            distances, indices = self.knn.kneighbors([row])
            distances = distances[0]  # Returns a list within a list so grab the inner list
            indices = indices[0]
            target_values = y[indices]

            # Grab the info for this observation
            my_id = self.df.iloc[my_index][self.id_column]
            my_output_data = self.df.iloc[my_index][output_columns]
            my_target = y[my_index]

            # Loop through the neighbors
            for n_index, n_distance, n_target in zip(indices, distances, target_values):
                # Skip myself
                if n_index == my_index:
                    continue

                # Compute the difference in feature space and target space
                feature_diff = n_distance
                target_diff = abs(my_target - n_target)

                # Compute target differences `within_distance` feature space
                if feature_diff <= within_distance and target_diff >= min_target_difference:
                    # Gather info about the neighbor
                    neighbor_id = self.df.iloc[n_index][self.id_column]
                    neighbor_output_data = self.df.iloc[n_index][output_columns]

                    # Add to the output DataFrame
                    row_data = my_output_data.to_dict()
                    row_data["feature_diff"] = feature_diff
                    row_data["target_diff"] = target_diff
                    row_data["n_id"] = neighbor_id
                    self.dataframe_builder.add_row(row_data)

                    # Print out the resolution issue (if verbose)
                    if verbose:
                        print(f"{output_count} Feature Diff: {feature_diff} Target Diff: {target_diff}")
                        print(f"\t{my_id}: {my_target:.3f} {list(my_output_data)}")
                        print(f"\t{neighbor_id}: {n_target:.3f} {list(neighbor_output_data)}")
                    # Increment the output count
                    output_count += 1

        # Return the output DataFrame
        return self.dataframe_builder.build()

    def recursive_compute(
        self, within_distance: float, min_target_difference: float, output_columns: list = [], verbose=True
    ) -> pd.DataFrame:
        """Compute Feature Resolution Issues, remove the issues, and recurse until no issues are found"""

        # Compute the resolution issues
        resolution_df = self.compute(within_distance, min_target_difference, output_columns, verbose)
        self.recursive_df_list.append(resolution_df)

        # If there are no resolution issues, return the combined DataFrame
        if len(resolution_df) == 0:
            return pd.concat(self.recursive_df_list)

        # Gather all IDs to be removed
        ids_to_remove = set(list(resolution_df[self.id_column]) + list(resolution_df["n_id"]))

        # Remove the rows of the observations that had issues
        print("Removing IDs: ", ids_to_remove)
        self.df = self.df[~self.df[self.id_column].isin(ids_to_remove)]

        # Recurse
        print("Recursing...")
        self.df = self.df.reset_index(drop=True)
        self.dataframe_builder = DataFrameBuilder()
        return self.recursive_compute(within_distance, min_target_difference, output_columns, verbose)


# Test the FeatureResolution Class
def simple_unit_test():
    """Test for the Feature Spider Class"""
    # Set some pandas options
    pd.set_option("display.max_columns", None)
    pd.set_option("display.width", 1000)

    # Make some fake data
    data = {
        "ID": [
            "id_0",
            "id_1",
            "id_2",
            "id_3",
            "id_4",
            "id_5",
            "id_6",
            "id_7",
            "id_8",
            "id_9",
        ],
        "feat1": [1.0, 1.0, 1.1, 3.0, 4.0, 1.0, 1.0, 1.1, 3.0, 4.0],
        "feat2": [1.0, 1.0, 1.1, 3.0, 4.0, 1.0, 1.0, 1.1, 3.0, 4.0],
        "feat3": [0.1, 0.2, 0.2, 1.6, 2.5, 0.1, 0.3, 0.2, 1.6, 2.5],
        "price": [10, 11, 12, 40, 20, 35, 61, 60, 40, 20],
    }
    data_df = pd.DataFrame(data)

    # Create the class and run the report
    resolution = FeatureResolution(data_df, features=["feat1", "feat2", "feat3"], target_column="price", id_column="ID")
    resolution.compute(within_distance=0.1, min_target_difference=10)


def unit_test():
    """Unit Test for the FeatureResolution Class"""
    from workbench.api.feature_set import FeatureSet
    from workbench.api.model import Model

    # Grab a test dataframe
    fs = FeatureSet("aqsol_mol_descriptors")
    test_df = fs.pull_dataframe()

    # Get the target and feature columns
    m = Model("aqsol-mol-regression")
    target_column = m.target()
    feature_columns = m.features()

    # Create the class and run the report
    resolution = FeatureResolution(
        test_df, features=feature_columns, target_column=target_column, id_column=fs.id_column
    )
    df = resolution.compute(within_distance=0.01, min_target_difference=1.0)
    print(df)


def recursive_test():
    from workbench.api.feature_set import FeatureSet
    from workbench.api.model import Model

    # Grab a test dataframe
    fs = FeatureSet("aqsol_mol_descriptors")
    test_df = fs.pull_dataframe()

    # Get the target and feature columns
    m = Model("aqsol-mol-regression")
    target_column = m.target()
    feature_columns = m.features()

    # Create the class and run the report
    resolution = FeatureResolution(
        test_df, features=feature_columns, target_column=target_column, id_column=fs.id_column
    )
    df = resolution.recursive_compute(within_distance=0.01, min_target_difference=1.0)
    print(df)


if __name__ == "__main__":
    simple_unit_test()
    unit_test()
    # recursive_test()



"""CachedArtifactMixin for caching method results in Artifact subclasses"""

import logging
import time
from functools import wraps
from concurrent.futures import ThreadPoolExecutor
from workbench.utils.workbench_cache import WorkbenchCache


class CachedArtifactMixin:
    """Mixin for caching methods in Artifact subclasses"""

    # Class-level caches, thread pool, and shutdown flag
    log = logging.getLogger("workbench")
    artifact_cache = WorkbenchCache(prefix="artifact_cache")
    fresh_cache = WorkbenchCache(prefix="artifact_fresh_cache", expire=120)
    thread_pool = ThreadPoolExecutor(max_workers=5)

    @staticmethod
    def _flatten_redis_key(method, *args, **kwargs):
        """Flatten the args and kwargs into a single string"""
        arg_str = "_".join(str(arg) for arg in args)
        kwarg_str = "_".join(f"{k}_{v}" for k, v in sorted(kwargs.items()))
        return f"{method.__name__}_{arg_str}_{kwarg_str}".replace(" ", "").replace("'", "")

    @classmethod
    def cache_result(cls, method):
        """Decorator to cache method results"""

        @wraps(method)
        def wrapper(self, *args, **kwargs):
            # Cache key includes the class name, instance Name, and method args/kwargs
            class_name = self.__class__.__name__.lower()
            cache_key = f"{class_name}_{self.name}_{cls._flatten_redis_key(method, *args, **kwargs)}"

            # Get the cached value and check if a refresh is needed
            cached_value = cls.artifact_cache.get(cache_key)
            cache_fresh = cls.fresh_cache.get(cache_key)

            # Check for the blocking case (no cached value)
            if cached_value is None:
                self.log.important(f"Blocking: Invoking method {method.__name__} for {cache_key}")
                cls.fresh_cache.set(cache_key, True)
                result = method(self, *args, **kwargs)
                cls.artifact_cache.set(cache_key, result)
                return result

            # Stale cache: Refresh in the background
            if cache_fresh is None:
                self.log.debug(f"Async: Refresh thread started: {cache_key}...")
                cls.fresh_cache.set(cache_key, True)
                cls.thread_pool.submit(cls._refresh_data_in_background, self, cache_key, method, *args, **kwargs)

            # Return the cached value (fresh or stale)
            return cached_value

        return wrapper

    @classmethod
    def _refresh_data_in_background(cls, instance, cache_key, method, *args, **kwargs):
        """Background data refresh method"""
        try:
            result = method(instance, *args, **kwargs)
            cls.artifact_cache.set(cache_key, result)
        except Exception as e:
            instance.log.error(f"Error refreshing data for {cache_key}: {e}")

    @classmethod
    def _shutdown(cls):
        """Explicitly shutdown the thread pool, if needed.
        Note: You should NOT call this method unless you know what you're doing."""
        if cls.thread_pool:
            cls.log.important("Shutting down the ThreadPoolExecutor...")
            time.sleep(10)
            try:
                cls.thread_pool.shutdown(wait=True)  # Gracefully shutdown
            except RuntimeError as e:
                cls.log.error(f"Error during thread pool shutdown: {e}")
            finally:
                cls.thread_pool = None


# Example usage of CachedModel
if __name__ == "__main__":
    from pprint import pprint
    from workbench.cached.cached_model import CachedModel

    my_model = CachedModel("abalone-regression")
    pprint(my_model.summary())
    pprint(my_model.details())
    # Second call to demonstrate caching
    pprint(my_model.summary())
    pprint(my_model.details())
    CachedArtifactMixin._shutdown()



# This code shows how workbench stores and retrieves tags from AWS
# You could use this code to retrieve tags from AWS and display them in a UI

import sagemaker
import base64
import json
from pprint import pprint

model_group_name = "abalone-regression"


# Helper method to decode tags
def decode_value(value):
    # Try to base64 decode the value
    try:
        value = base64.b64decode(value).decode("utf-8")
    except Exception:
        pass
    # Try to JSON decode the value
    try:
        value = json.loads(value)
    except Exception:
        pass

    # Okay, just return whatever we have
    return value


# Helper method to stitch together tags
def aws_tags_to_dict(aws_tags) -> dict:
    """Internal: AWS Tags are in an odd format, so convert, decode, and de-chunk"""

    # Stitch together any chunked data
    stitched_data = {}
    regular_tags = {}

    for item in aws_tags:
        key = item["Key"]
        value = item["Value"]

        # Check if this key is a chunk
        if "_chunk_" in key:
            base_key, chunk_num = key.rsplit("_chunk_", 1)

            if base_key not in stitched_data:
                stitched_data[base_key] = {}

            stitched_data[base_key][int(chunk_num)] = value
        else:
            regular_tags[key] = decode_value(value)

    # Stitch chunks back together and decode
    for base_key, chunks in stitched_data.items():
        # Sort by chunk number and concatenate
        sorted_chunks = [chunks[i] for i in sorted(chunks.keys())]
        stitched_base64_str = "".join(sorted_chunks)

        # Decode the stitched base64 string
        try:
            stitched_json_str = base64.b64decode(stitched_base64_str).decode("utf-8")
        except UnicodeDecodeError:
            stitched_json_str = stitched_base64_str
        try:
            stitched_dict = json.loads(stitched_json_str)
        except json.decoder.JSONDecodeError:
            stitched_dict = stitched_json_str

        regular_tags[base_key] = stitched_dict

    return regular_tags


sm_session = sagemaker.Session()
client = sm_session.sagemaker_client

# Retrieve the model package group ARN
response = client.describe_model_package_group(ModelPackageGroupName=model_group_name)
group_arn = response["ModelPackageGroupArn"]

# Use the SageMaker session's list_tags method
tags = sm_session.list_tags(resource_arn=group_arn)

# Convert tags to dictionary
tags = aws_tags_to_dict(tags)
pprint(tags)



"""This Script creates the Classification Artifacts in AWS/Workbench

DataSources:
    - wine_data_copy
FeatureSets:
    - wine_feature_set_copy
Models:
    - wine-classification-copy
Endpoints:
    - wine-classification-copy
"""

import sys

from pathlib import Path
from workbench.api import DataSource, FeatureSet, Model, ModelType, Endpoint


if __name__ == "__main__":

    # Get the path to the dataset in the repository data directory
    wine_data_copy_path = Path(sys.modules["workbench"].__file__).parent.parent.parent / "data" / "wine_dataset.csv"

    # Recreate Flag in case you want to recreate the artifacts
    recreate = False

    # Create the wine_data_copy DataSource
    if recreate or not DataSource("wine_data_copy").exists():
        DataSource(wine_data_copy_path, name="wine_data_copy")

    # Create the wine_features_copy FeatureSet
    if recreate or not FeatureSet("wine_features_copy").exists():
        ds = DataSource("wine_data_copy")
        ds.to_features("wine_features_copy", tags=["wine", "classification"])

    # Create the wine classification Model
    if recreate or not Model("wine-classification-copy").exists():
        fs = FeatureSet("wine_features_copy")
        m = fs.to_model(
            name="wine-classification-copy",
            model_type=ModelType.CLASSIFIER,
            target_column="wine_class",
            tags=["wine", "classification"],
            description="Wine Classification Model",
        )
        m.set_owner("test")

    # Create the wine classification Endpoint
    if recreate or not Endpoint("wine-classification-copy").exists():
        m = Model("wine-classification-copy")
        end = m.to_endpoint("wine-classification-copy", tags=["wine", "classification"])

        # Run inference on the endpoint
        end.auto_inference(capture=True)



"""DataToDataExample: Example Class that demonstrates a Data Source to Data Source Transform"""

# Local imports
from workbench.core.transforms.data_to_data.light.data_to_data_light import DataToDataLight
from workbench.utils import pandas_utils


class DataToDataExample(DataToDataLight):
    def __init__(self, input_name: str, output_name: str):
        """DataToDataExample: Example Class that demonstrates a Data Source to Data Source Transform"""

        # Call superclass init
        super().__init__(input_name, output_name)

    def transform_impl(self, drop_na="any"):
        """Just dropping NaNs, but you could do anything that you want, simply take the
        input dataframe and produce an output dataframe (of any form)"""

        # Drop Rows that have NaNs in them
        self.output_df = pandas_utils.remove_rows_with_nans(self.input_df, how=drop_na)


# Simple test of the DataToDataExample functionality
def test():
    """Test the DataToDataExample Class"""

    # Create the class with inputs and outputs and invoke the transform
    input_name = "test_data"
    output_name = "test_data_clean"
    DataToDataExample(input_name, output_name).transform(drop_na="any")


if __name__ == "__main__":
    test()



# Description: Give a report on the redis server data size and last modified time.
import argparse
import logging
from workbench.utils.redis_cache import RedisCache

# Set up logging
log = logging.getLogger("workbench")


def redis_report(n: int = 20):
    """Print the top N Redis Data sizes and last modified time."""
    redis_cache = RedisCache()

    # Check if Redis is available
    if not redis_cache.check():
        log.error("Redis not available, exiting...")
        return

    # Report memory usage and configuration settings
    redis_cache.report_memory_config()

    # Report the largest keys in the Redis database
    redis_cache.report_largest_keys(n)


def parse_args():
    parser = argparse.ArgumentParser(description="Redis Report.")
    parser.add_argument("--n", type=int, default=20, help="Top N Redis Data sizes Default is 20.")
    return parser.parse_args()


def main():
    args = parse_args()
    redis_report(args.n)


if __name__ == "__main__":
    main()



"""Monitor: Manages AWS Endpoint Monitor creation and deployment.
Endpoints Monitors are set up and provisioned for deployment into AWS.
Monitors can be viewed in the AWS Sagemaker interfaces or in the Workbench
Dashboard UI, which provides additional monitor details and performance metrics"""

import pandas as pd
from typing import Union

# Workbench Imports
from workbench.core.artifacts.monitor_core import MonitorCore


class Monitor(MonitorCore):
    """Monitor: Workbench Monitor API Class

    Common Usage:
       ```
       mon = Endpoint(name).get_monitor()  # Pull from endpoint OR
       mon = Monitor(name)                 # Create using Endpoint Name
       mon.summary()
       mon.details()

       # One time setup methods
       mon.enable_data_capture()
       mon.create_baseline()
       mon.create_monitoring_schedule()

       # Pull information from the monitor
       baseline_df = mon.get_baseline()
       constraints_df = mon.get_constraints()
       stats_df = mon.get_statistics()
       input_df, output_df = mon.get_captured_data()
       ```
    """

    def summary(self) -> dict:
        """Monitor Summary

        Returns:
            dict: A dictionary of summary information about the Monitor
        """
        return super().summary()

    def details(self) -> dict:
        """Monitor Details

        Returns:
            dict: A dictionary of details about the Monitor
        """
        return super().details()

    def enable_data_capture(self, capture_percentage=100):
        """
        Enable data capture configuration for this Monitor/endpoint.

        Args:
            capture_percentage (int): Percentage of data to capture. Defaults to 100.
        """
        super().enable_data_capture(capture_percentage)

    def create_baseline(self, recreate: bool = False):
        """Code to create a baseline for monitoring

        Args:
            recreate (bool): If True, recreate the baseline even if it already exists

        Notes:
            This will create/write three files to the baseline_dir:
            - baseline.csv
            - constraints.json
            - statistics.json
        """
        super().create_baseline(recreate)

    def create_monitoring_schedule(self, schedule: str = "hourly"):
        """
        Sets up the monitoring schedule for the model endpoint.

        Args:
            schedule (str): The schedule for the monitoring job (hourly or daily, defaults to hourly).
        """
        super().create_monitoring_schedule(schedule)

    def get_captured_data(self) -> (pd.DataFrame, pd.DataFrame):
        """
        Get the latest data capture input and output from S3.

        Returns:
            DataFrame (input), DataFrame(output): Flattened and processed DataFrames for input and output data.
        """
        return super().get_captured_data()

    def get_baseline(self) -> Union[pd.DataFrame, None]:
        """Code to get the baseline CSV from the S3 baseline directory

        Returns:
            pd.DataFrame: The baseline CSV as a DataFrame (None if it doesn't exist)
        """
        return super().get_baseline()

    def get_constraints(self) -> Union[pd.DataFrame, None]:
        """Code to get the constraints from the baseline

        Returns:
           pd.DataFrame: The constraints from the baseline (constraints.json) (None if it doesn't exist)
        """
        return super().get_constraints()

    def get_statistics(self) -> Union[pd.DataFrame, None]:
        """Code to get the statistics from the baseline

        Returns:
            pd.DataFrame: The statistics from the baseline (statistics.json) (None if it doesn't exist)
        """
        return super().get_statistics()


if __name__ == "__main__":
    """Exercise the MonitorCore class"""
    from pprint import pprint
    from workbench.api.endpoint import Endpoint

    # Set options for actually seeing the dataframe
    pd.set_option("display.max_columns", None)
    pd.set_option("display.width", None)

    # Create the Class and test it out
    endpoint_name = "abalone-regression-rt"
    my_endpoint = Endpoint(endpoint_name)
    if not my_endpoint.exists():
        print(f"Endpoint {endpoint_name} does not exist.")
        exit(1)
    mm = Monitor(endpoint_name)

    # Check the summary and details of the monitoring class
    pprint(mm.summary())
    pprint(mm.details())

    # Enable data capture for this endpoint
    mm.enable_data_capture()

    # Create a baseline for monitoring
    mm.create_baseline()

    # Set up the monitoring schedule (if it doesn't already exist)
    mm.create_monitoring_schedule()

    # Check the monitoring outputs
    print("\nBaseline outputs...")
    base_df = mm.get_baseline()
    print(base_df.head())

    print("\nConstraints...")
    pprint(mm.get_constraints())

    print("\nStatistics...")
    print(mm.get_statistics())

    # Get the latest data capture
    input_df, output_df = mm.get_captured_data()
    print(input_df.head())
    print(output_df.head())



# Workbench Imports
from workbench.api.df_store import DFStore
from workbench.utils.chem_utils import compute_morgan_fingerprints
from rdkit.Chem import DataStructs
from rdkit.DataStructs.cDataStructs import CreateFromBitString
import pandas as pd

# Grab tox21 dataset (just 10)
tox_df = DFStore().get("/datasets/chem_info/tox21_10")

# Compute Morgan fingerprints as bitstrings
tox_df = compute_morgan_fingerprints(tox_df)

# Extract fingerprints and IDs
fingerprint_column = "fingerprint"
id_column = "id"
fingerprints = tox_df[fingerprint_column].tolist()
compound_ids = tox_df[id_column].tolist()

# Convert bitstrings to ExplicitBitVect objects
explicit_fingerprints = [CreateFromBitString(fp) for fp in fingerprints]

# Compute Tanimoto similarities using RDKit
results = []
for i, query_fp in enumerate(explicit_fingerprints):
    # Compute bulk Tanimoto similarity
    similarities = DataStructs.BulkTanimotoSimilarity(query_fp, explicit_fingerprints)
    # Get top 5 neighbors (excluding the query itself)
    top_indices = sorted(range(len(similarities)), key=lambda j: -similarities[j])[:6]  # Include self
    top_neighbors = [
        {"query_id": compound_ids[i], "neighbor_id": compound_ids[j], "similarity": similarities[j]}
        for j in top_indices
        if compound_ids[j] != compound_ids[i]
    ][
        :5
    ]  # Exclude self and limit to top 5
    results.extend(top_neighbors)

# Convert results to a DataFrame
neighbors_df = pd.DataFrame(results)

# Display all neighbors for all compounds
print("\nAll neighbors:")
print(neighbors_df.head(20))

# Query for neighbors for a specific compound
query_id = tox_df[id_column].iloc[1]
query_neighbors_df = neighbors_df[neighbors_df["query_id"] == query_id]
print(f"\nBULK: Neighbors for query ID {query_id}:")
print(query_neighbors_df)



import os
import boto3
import logging
import aws_cdk as cdk
from workbench_core.workbench_core_stack import WorkbenchCoreStack, WorkbenchCoreStackProps

# Initialize ConfigManager with error handling
try:
    # Temporarily disable logging
    # logging.disable(logging.CRITICAL)
    from workbench.utils.config_manager import ConfigManager

    cm = ConfigManager()
except Exception as e:
    print(f"Workbench ConfigManager initialization failed: {e}")
    print("Falling back to environment variables only")
    cm = None
finally:
    # Re-enable logging
    logging.disable(logging.NOTSET)

# Grab the account and region using boto3
session = boto3.session.Session()
aws_account = session.client("sts").get_caller_identity().get("Account")
aws_region = session.region_name
print(f"Account: {aws_account}")
print(f"Region: {aws_region}")

# Get configurations for Workbench Bucket, SSO Group, and Additional Buckets
workbench_bucket = (cm and cm.get_config("WORKBENCH_BUCKET")) or os.getenv("WORKBENCH_BUCKET")
if not workbench_bucket:
    print("Error: WORKBENCH_BUCKET is required but not found in config or environment variables.")
    exit(1)

# SSO Groups (Groups that can assume the Workbench roles)
sso_groups_str = (cm and cm.get_config("WORKBENCH_SSO_GROUPS")) or os.getenv("WORKBENCH_SSO_GROUPS", "")
sso_groups = [group.strip() for group in sso_groups_str.split(",") if group.strip()] if sso_groups_str else []

# Additional Buckets (Extra S3 buckets to grant access to)
additional_buckets_str = (cm and cm.get_config("WORKBENCH_ADDITIONAL_BUCKETS")) or os.getenv(
    "WORKBENCH_ADDITIONAL_BUCKETS", ""
)
additional_buckets = (
    [bucket.strip() for bucket in additional_buckets_str.split(",") if bucket.strip()] if additional_buckets_str else []
)
existing_vpc_id = (cm and cm.get_config("WORKBENCH_VPC_ID")) or os.getenv("WORKBENCH_VPC_ID")
subnet_ids = (cm and cm.get_config("WORKBENCH_SUBNET_IDS")) or os.getenv("WORKBENCH_SUBNET_IDS", "")

# Log the configuration for transparency
print("Configuration:")
print(f"  WORKBENCH_BUCKET: {workbench_bucket}")
print(f"  WORKBENCH_SSO_GROUPS: {sso_groups}")
print(f"  WORKBENCH_ADDITIONAL_BUCKETS: {additional_buckets}")
print(f"  WORKBENCH_VPC_ID: {existing_vpc_id}")
print(f"  WORKBENCH_SUBNET_IDS: {subnet_ids}")

# Our CDK App and Environment
app = cdk.App()
env = cdk.Environment(account=aws_account, region=aws_region)

# Create the Workbench Core Stack
sandbox_stack = WorkbenchCoreStack(
    app,
    "WorkbenchCore",
    env=env,
    props=WorkbenchCoreStackProps(
        workbench_bucket=workbench_bucket,
        sso_groups=sso_groups,
        additional_buckets=additional_buckets,
        existing_vpc_id=existing_vpc_id,
        subnet_ids=subnet_ids,
    ),
)

app.synth()



"""CachedMeta: A class that provides caching for the Meta() class"""

import logging
from typing import Union
import pandas as pd
from functools import wraps
from concurrent.futures import ThreadPoolExecutor


# Workbench Imports
from workbench.core.cloud_platform.cloud_meta import CloudMeta
from workbench.utils.workbench_cache import WorkbenchCache


# Decorator to cache method results from the Meta class
def cache_result(method):
    """Decorator to cache method results in meta_cache"""

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        # Create a unique cache key based on the method name and arguments
        cache_key = CachedMeta._flatten_redis_key(method, *args, **kwargs)

        # Check for fresh data, spawn thread to refresh if stale
        if self.fresh_cache.atomic_set(cache_key, True):
            self.log.important(f"Async: Metadata for {cache_key} refresh thread started...")
            self.thread_pool.submit(self._refresh_data_in_background, cache_key, method, *args, **kwargs)

        # Return data (fresh or stale) if available
        cached_value = self.meta_cache.get(cache_key)
        if cached_value is not None:
            return cached_value

        # Fall back to calling the method if no cached data found
        self.log.important(f"Blocking: Getting Metadata for {cache_key}")
        result = method(self, *args, **kwargs)
        self.meta_cache.set(cache_key, result)
        return result

    return wrapper


class CachedMeta(CloudMeta):
    """CachedMeta: Singleton class for caching metadata functionality.

    Common Usage:
       ```python
       from workbench.cached.cached_meta import CachedMeta
       meta = CachedMeta()

       # Get the AWS Account Info
       meta.account()
       meta.config()

       # These are 'list' methods
       meta.etl_jobs()
       meta.data_sources()
       meta.feature_sets(details=True/False)
       meta.models(details=True/False)
       meta.endpoints(details=True/False)
       meta.views()

       # These are 'describe' methods
       meta.data_source("abalone_data")
       meta.feature_set("abalone_features")
       meta.model("abalone-regression")
       meta.endpoint("abalone-endpoint")
       ```
    """

    _instance = None  # Class attribute to hold the singleton instance

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(CachedMeta, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        """CachedMeta Initialization"""
        if hasattr(self, "_initialized") and self._initialized:
            return  # Prevent reinitialization

        self.log = logging.getLogger("workbench")
        self.log.important("Initializing CachedMeta...")
        super().__init__()

        # Create both our Meta Cache and Fresh Cache (tracks if data is stale)
        self.meta_cache = WorkbenchCache(prefix="meta")
        self.fresh_cache = WorkbenchCache(prefix="meta_fresh", expire=300)  # 5-minute expiration

        # Create a ThreadPoolExecutor for refreshing stale data
        self.thread_pool = ThreadPoolExecutor(max_workers=5)

        # Mark the instance as initialized
        self._initialized = True

    def check(self):
        """Check if our underlying caches are working"""
        return self.meta_cache.check()

    def list_meta_cache(self):
        """List the current Meta Cache"""
        return self.meta_cache.list_keys()

    def clear_meta_cache(self):
        """Clear the current Meta Cache"""
        self.meta_cache.clear()

    @cache_result
    def account(self) -> dict:
        """Cloud Platform Account Info

        Returns:
            dict: Cloud Platform Account Info
        """
        return super().account()

    @cache_result
    def config(self) -> dict:
        """Return the current Workbench Configuration

        Returns:
            dict: The current Workbench Configuration
        """
        return super().config()

    @cache_result
    def incoming_data(self) -> pd.DataFrame:
        """Get summary data about data in the incoming raw data

        Returns:
            pd.DataFrame: A summary of the incoming raw data
        """
        return super().incoming_data()

    @cache_result
    def etl_jobs(self) -> pd.DataFrame:
        """Get summary data about Extract, Transform, Load (ETL) Jobs

        Returns:
            pd.DataFrame: A summary of the ETL Jobs deployed in the Cloud Platform
        """
        return super().etl_jobs()

    @cache_result
    def data_sources(self) -> pd.DataFrame:
        """Get a summary of the Data Sources deployed in the Cloud Platform

        Returns:
            pd.DataFrame: A summary of the Data Sources deployed in the Cloud Platform
        """
        return super().data_sources()

    @cache_result
    def views(self, database: str = "workbench") -> pd.DataFrame:
        """Get a summary of the all the Views, for the given database, in AWS

        Args:
            database (str, optional): Glue database. Defaults to 'workbench'.

        Returns:
            pd.DataFrame: A summary of all the Views, for the given database, in AWS
        """
        return super().views(database=database)

    @cache_result
    def feature_sets(self, details: bool = False) -> pd.DataFrame:
        """Get a summary of the Feature Sets deployed in the Cloud Platform

        Args:
            details (bool, optional): Include detailed information. Defaults to False.

        Returns:
            pd.DataFrame: A summary of the Feature Sets deployed in the Cloud Platform
        """
        return super().feature_sets(details=details)

    @cache_result
    def models(self, details: bool = False) -> pd.DataFrame:
        """Get a summary of the Models deployed in the Cloud Platform

        Args:
            details (bool, optional): Include detailed information. Defaults to False.

        Returns:
            pd.DataFrame: A summary of the Models deployed in the Cloud Platform
        """
        return super().models(details=details)

    @cache_result
    def endpoints(self, details: bool = False) -> pd.DataFrame:
        """Get a summary of the Endpoints deployed in the Cloud Platform

        Args:
            details (bool, optional): Include detailed information. Defaults to False.

        Returns:
            pd.DataFrame: A summary of the Endpoints in the Cloud Platform
        """
        return super().endpoints(details=details)

    @cache_result
    def glue_job(self, job_name: str) -> Union[dict, None]:
        """Get the details of a specific Glue Job

        Args:
            job_name (str): The name of the Glue Job

        Returns:
            dict: The details of the Glue Job (None if not found)
        """
        return super().glue_job(job_name=job_name)

    @cache_result
    def data_source(self, data_source_name: str, database: str = "workbench") -> Union[dict, None]:
        """Get the details of a specific Data Source

        Args:
            data_source_name (str): The name of the Data Source
            database (str, optional): The Glue database. Defaults to 'workbench'.

        Returns:
            dict: The details of the Data Source (None if not found)
        """
        return super().data_source(data_source_name=data_source_name, database=database)

    @cache_result
    def feature_set(self, feature_set_name: str) -> Union[dict, None]:
        """Get the details of a specific Feature Set

        Args:
            feature_set_name (str): The name of the Feature Set

        Returns:
            dict: The details of the Feature Set (None if not found)
        """
        return super().feature_set(feature_set_name=feature_set_name)

    @cache_result
    def model(self, model_name: str) -> Union[dict, None]:
        """Get the details of a specific Model

        Args:
            model_name (str): The name of the Model

        Returns:
            dict: The details of the Model (None if not found)
        """
        return super().model(model_name=model_name)

    @cache_result
    def endpoint(self, endpoint_name: str) -> Union[dict, None]:
        """Get the details of a specific Endpoint

        Args:
            endpoint_name (str): The name of the Endpoint

        Returns:
            dict: The details of the Endpoint (None if not found)
        """
        return super().endpoint(endpoint_name=endpoint_name)

    def _refresh_data_in_background(self, cache_key, method, *args, **kwargs):
        """Background task to refresh AWS metadata."""
        result = method(self, *args, **kwargs)
        self.meta_cache.set(cache_key, result)
        self.log.important(f"Updated Metadata for {cache_key}")

    @staticmethod
    def _flatten_redis_key(method, *args, **kwargs):
        """Flatten the args and kwargs into a single string"""
        arg_str = "_".join(str(arg) for arg in args)
        kwarg_str = "_".join(f"{k}_{v}" for k, v in sorted(kwargs.items()))
        return f"{method.__name__}_{arg_str}_{kwarg_str}".replace(" ", "").replace("'", "")

    def __del__(self):
        """Destructor to shut down the thread pool gracefully."""
        self.close()

    def close(self):
        """Explicitly close the thread pool, if needed."""
        if self.thread_pool:
            self.log.important("Shutting down the ThreadPoolExecutor...")
            try:
                self.thread_pool.shutdown(wait=True)  # Gracefully shutdown
            except RuntimeError as e:
                self.log.error(f"Error during thread pool shutdown: {e}")
            finally:
                self.thread_pool = None

    def __repr__(self):
        return f"CachedMeta()\n\t{repr(self.meta_cache)}\n\t{super().__repr__()}"


if __name__ == "__main__":
    """Exercise the Workbench AWSCachedMeta Class"""
    from pprint import pprint
    import time

    # Pandas Display Options
    pd.set_option("display.max_columns", None)
    pd.set_option("display.width", 1000)

    # Create the class
    meta = CachedMeta()
    print(f"Check: {meta.check()}")

    # Test the __repr__ method
    print(meta)

    # List the current Meta Cache
    meta.list_meta_cache()

    # Clear the current Meta Cache
    # meta.clear_meta_cache()

    # Get the AWS Account Info
    print("*** AWS Account ***")
    pprint(meta.account())

    # Get the Workbench Configuration
    print("*** Workbench Configuration ***")
    pprint(meta.config())

    # Get the Incoming Data
    print("\n\n*** Incoming Data ***")
    print(meta.incoming_data())

    # Get the AWS Glue Jobs (ETL Jobs)
    print("\n\n*** ETL Jobs ***")
    print(meta.etl_jobs())

    # Get the Data Sources
    print("\n\n*** Data Sources ***")
    print(meta.data_sources())

    # Get the Views (Data Sources)
    print("\n\n*** Views (Data Sources) ***")
    print(meta.views("workbench"))

    # Get the Views (Feature Sets)
    print("\n\n*** Views (Feature Sets) ***")
    fs_views = meta.views("sagemaker_featurestore")
    print(fs_views)

    # Get the Feature Sets
    print("\n\n*** Feature Sets ***")
    pprint(meta.feature_sets())

    # Get the Models
    print("\n\n*** Models ***")
    start_time = time.time()
    pprint(meta.models())
    print(f"Elapsed Time Model (no details): {time.time() - start_time:.2f}")

    # Get the Models with Details
    print("\n\n*** Models with Details ***")
    start_time = time.time()
    pprint(meta.models(details=True))
    print(f"Elapsed Time Model (with details): {time.time() - start_time:.2f}")

    # Get the Endpoints
    print("\n\n*** Endpoints ***")
    pprint(meta.endpoints())

    # Test out the specific artifact details methods
    print("\n\n*** Glue Job Details ***")
    pprint(meta.glue_job("Glue_Job_1"))
    print("\n\n*** DataSource Details ***")
    pprint(meta.data_source("abalone_data"))
    print("\n\n*** FeatureSet Details ***")
    pprint(meta.feature_set("abalone_features"))
    print("\n\n*** Model Details ***")
    pprint(meta.model("abalone-regression"))
    print("\n\n*** Endpoint Details ***")
    pprint(meta.endpoint("abalone-regression"))

    # Test out a non-existent model
    print("\n\n*** Model Doesn't Exist ***")
    pprint(meta.model("non-existent-model"))



"""PipelinesPageView pulls Pipeline metadata from the AWS Service Broker with Details Panels on each Pipeline"""

import pandas as pd

# Workbench Imports
from workbench.web_interface.page_views.page_view import PageView
from workbench.cached.cached_meta import CachedMeta
from workbench.cached.cached_pipeline import CachedPipeline
from workbench.utils.symbols import tag_symbols


class PipelinesPageView(PageView):
    def __init__(self):
        """PipelinesPageView pulls Pipeline metadata and populates a Details Panel"""
        # Call SuperClass Initialization
        super().__init__()

        # CachedMeta object for Cloud Platform Metadata
        self.meta = CachedMeta()

        # Initialize the Pipelines DataFrame
        self.pipelines_df = None
        self.refresh()

    def refresh(self):
        """Refresh the pipeline data from the Cloud Platform"""
        self.log.important("Calling pipelines page view refresh()..")
        self.pipelines_df = self.meta.pipelines()

        # Drop the AWS URL column
        self.pipelines_df.drop(columns=["_aws_url"], inplace=True, errors="ignore")
        # Add Health Symbols to the Model Group Name
        if "Health" in self.pipelines_df.columns:
            self.pipelines_df["Health"] = self.pipelines_df["Health"].map(lambda x: tag_symbols(x))

    def pipelines(self) -> pd.DataFrame:
        """Get all the data that's useful for this view

        Returns:
            pd.DataFrame: DataFrame of the Pipelines View Data
        """
        return self.pipelines_df

    @staticmethod
    def pipeline_details(pipeline_name: str) -> (dict, None):
        """Get all the details for the given Pipeline Name
         Args:
            pipeline_name(str): The Name of the Pipeline
        Returns:
            dict: The details for the given Model (or None if not found)
        """
        pipeline = CachedPipeline(pipeline_name)
        if pipeline is None:
            return {"Status": "Not Found"}

        # Return the Pipeline Details
        return pipeline.details()


if __name__ == "__main__":
    # Exercising the PipelinesPageView
    import time
    from pprint import pprint

    # Create the class and get the AWS Pipeline details
    pipeline_view = PipelinesPageView()

    # List the Pipelines
    print("PipelinesSummary:")
    summary = pipeline_view.pipelines()
    print(summary.head())

    # Get the details for the first Pipeline
    my_pipeline_name = summary["Name"].iloc[0]
    print("\nPipelineDetails:")
    details = pipeline_view.pipeline_details(my_pipeline_name)
    pprint(details)

    # Give any broker threads time to finish
    time.sleep(1)



"""Example script of running inference on an Endpoint"""

import argparse
import pandas as pd

# Workbench Imports
from workbench.core.artifacts.feature_set_core import FeatureSetCore
from workbench.core.artifacts.model_core import ModelCore
from workbench.core.artifacts.endpoint_core import EndpointCore


def run_inference(endpoint_name):
    # Set options for actually seeing the dataframe
    pd.set_option("display.max_columns", None)
    pd.set_option("display.width", None)

    # Grab the Endpoint
    my_endpoint = EndpointCore(endpoint_name)

    # Grab the FeatureSet by backtracking from the Endpoint
    model = my_endpoint.get_input()
    feature_set = ModelCore(model).get_input()
    features = FeatureSetCore(feature_set)
    table = features.view("training").table
    test_df = features.query(f'SELECT * FROM "{table}" where training = FALSE')

    # Drop some columns
    test_df.drop(["write_time", "api_invocation_time", "is_deleted"], axis=1, inplace=True)

    # Make predictions on the Endpoint
    pred_df = my_endpoint.predict(test_df[:10])
    print(pred_df.head())


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run inference on a Workbench Endpoint")
    parser.add_argument("endpoint_name", type=str, help="Name of the Workbench Endpoint")
    args = parser.parse_args()

    run_inference(args.endpoint_name)



"""Welcome to the Workbench Pandas Transform Classes

These classes provide low-level APIs for using Pandas DataFrames

- DataToPandas: Pull a dataframe from a Workbench DataSource
- FeaturesToPandas: Pull a dataframe from a Workbench FeatureSet
- PandasToData: Create a Workbench DataSource using a Pandas DataFrame as the source
- PandasToFeatures: Create a Workbench FeatureSet using a Pandas DataFrame as the source
- PandasToFeaturesChunked: Create a Workbench FeatureSet using a Chunked/Streaming Pandas DataFrame as the source
"""

from .data_to_pandas import DataToPandas
from .features_to_pandas import FeaturesToPandas
from .pandas_to_data import PandasToData
from .pandas_to_features import PandasToFeatures
from .pandas_to_features_chunked import PandasToFeaturesChunked

__all__ = ["DataToPandas", "FeaturesToPandas", "PandasToData", "PandasToFeatures", "PandasToFeaturesChunked"]



import os
import pandas as pd
import numpy as np
from sklearn.preprocessing import LabelEncoder
from sklearn.metrics import (
    precision_recall_fscore_support,
    confusion_matrix,
    mean_absolute_error,
    r2_score,
    root_mean_squared_error,
)

# Import pytorch-tabular components
# Set this before importing any PyTorch-related modules to fix weight loading issues
os.environ["TORCH_FORCE_NO_WEIGHTS_ONLY_LOAD"] = "1"
import pytest  # noqa: E402

pytest.skip("skipping this entire module for now", allow_module_level=True)
from pytorch_tabular import TabularModel  # noqa: E402
from pytorch_tabular.models import TabNetModelConfig, CategoryEmbeddingModelConfig  # noqa: E402
from pytorch_tabular.config import DataConfig, OptimizerConfig, TrainerConfig  # noqa: E402


def process_predictions(result, target_name, label_encoder=None, model_type="classifier"):
    """
    Process the prediction results from pytorch-tabular
    """
    df_result = result.copy()

    if model_type == "classifier":
        # Get predictions from {target}_prediction column
        pred_col = f"{target_name}_prediction"
        preds = df_result[pred_col].values

        # Get probability columns - they should be {target}_0_probability, {target}_1_probability, etc.
        prob_cols = [
            col for col in df_result.columns if col.startswith(f"{target_name}_") and col.endswith("_probability")
        ]
        prob_cols.sort()  # Ensure consistent ordering

        if prob_cols and label_encoder:
            # Create a list of probabilities for each row
            probs = df_result[prob_cols].values
            df_result = df_result.copy()
            df_result["pred_proba"] = [p.tolist() for p in probs]

            # Also create named columns for each class
            for i, class_name in enumerate(label_encoder.classes_):
                if i < len(prob_cols):
                    df_result[f"{class_name}_proba"] = df_result[prob_cols[i]]

        # Decode predictions if we have a label encoder
        if label_encoder:
            preds_decoded = label_encoder.inverse_transform(preds.astype(int))
        else:
            preds_decoded = preds

        return df_result, preds_decoded

    else:  # regression
        # For regression, look for {target}_prediction
        pred_col = f"{target_name}_prediction"
        if pred_col in df_result.columns:
            preds = df_result[pred_col].values
        else:
            # Fallback to target column name
            preds = df_result[target_name].values

        return df_result, preds


def create_sample_data(model_type="classifier", n_samples=5000):
    """
    Create sample training and validation data
    """
    np.random.seed(42)

    # Create sample features
    data = {
        "feature1": np.random.randn(n_samples),
        "feature2": np.random.randn(n_samples),
        "feature3": np.random.choice(["A", "B", "C"], n_samples),
        "feature4": np.random.randint(0, 100, n_samples).astype(float),
    }

    if model_type == "classifier":
        # Create binary classification target
        data["target"] = np.random.choice(["Class1", "Class2"], n_samples)
    else:
        # Create regression target
        data["target"] = 2 * data["feature1"] + 1.5 * data["feature2"] + np.random.randn(n_samples) * 0.1

    df = pd.DataFrame(data)

    # Split into train/validation
    split_idx = int(0.8 * len(df))
    df_train = df[:split_idx].copy()
    df_val = df[split_idx:].copy()

    return df_train, df_val


def test_model_training(model_type="classifier"):
    """
    Test the model training pipeline
    """
    print(f"Testing {model_type} model...")

    # Create sample data
    df_train, df_val = create_sample_data(model_type)
    target = "target"

    # Configure data settings - minimal required config
    continuous_cols = ["feature1", "feature2", "feature4"]
    categorical_cols = ["feature3"]

    data_config = DataConfig(
        target=[target],
        continuous_cols=continuous_cols,
        categorical_cols=categorical_cols,
    )

    # Use default configs - only override what we need
    optimizer_config = OptimizerConfig()
    trainer_config = TrainerConfig(
        auto_lr_find=False,  # Disable to avoid issues with small dataset
        max_epochs=10,  # Just set epochs, use defaults for everything else
    )

    # Choose model configuration based on model type
    if model_type == "classifier":
        task = "classification"
        model_config = TabNetModelConfig(task=task)
        # Encode the target column - create new dataframes to avoid chained assignment
        label_encoder = LabelEncoder()

        # Create completely fresh copies and assign the target column properly
        df_train_encoded = df_train.copy()
        df_val_encoded = df_val.copy()

        # Use proper assignment to avoid pandas warnings
        target_train_encoded = label_encoder.fit_transform(df_train[target])
        target_val_encoded = label_encoder.transform(df_val[target])

        df_train_encoded[target] = target_train_encoded
        df_val_encoded[target] = target_val_encoded
    else:
        task = "regression"
        model_config = CategoryEmbeddingModelConfig(task=task)
        label_encoder = None
        df_train_encoded = df_train
        df_val_encoded = df_val
    # Create and train the TabularModel
    tabular_model = TabularModel(
        data_config=data_config,
        model_config=model_config,
        optimizer_config=optimizer_config,
        trainer_config=trainer_config,
    )
    # Train the model
    print("Training model...")
    tabular_model.fit(train=df_train_encoded, validation=df_val_encoded)
    # Make Predictions on the Validation Set
    print("Making Predictions on Validation Set...")
    result = tabular_model.predict(df_val_encoded)

    print("Available columns in result:", result.columns.tolist())
    # Process predictions
    df_result, preds = process_predictions(result, target, label_encoder, model_type)

    # Get actual target values for comparison
    if model_type == "classifier":
        y_validate = label_encoder.inverse_transform(df_val_encoded[target])
    else:
        y_validate = df_val_encoded[target].values
    # Add the decoded predictions to the result dataframe
    df_result = df_result.copy()
    df_result["prediction"] = preds

    # Define output columns based on what we actually have
    output_columns = []

    # Always include target and prediction
    if target in df_result.columns:
        output_columns.append(target)
    output_columns.append("prediction")

    # Add probability columns if they exist
    if model_type == "classifier":
        # Add the pytorch-tabular native probability columns
        prob_cols = [col for col in df_result.columns if col.startswith(f"{target}_") and col.endswith("_probability")]
        output_columns.extend(sorted(prob_cols))

        # Add our custom named probability columns
        class_prob_cols = [col for col in df_result.columns if col.endswith("_proba") and not col == "pred_proba"]
        output_columns.extend(sorted(class_prob_cols))

        # Add the list version if it exists
        if "pred_proba" in df_result.columns:
            output_columns.append("pred_proba")

    # Report Performance Metrics
    if model_type == "classifier":
        label_names = label_encoder.classes_
        scores = precision_recall_fscore_support(y_validate, preds, average=None, labels=label_names)
        score_df = pd.DataFrame(
            {
                target: label_names,
                "precision": scores[0],
                "recall": scores[1],
                "fscore": scores[2],
                "support": scores[3],
            }
        )
        print("\nClassification Metrics:")
        metrics = ["precision", "recall", "fscore", "support"]
        for t in label_names:
            for m in metrics:
                value = score_df.loc[score_df[target] == t, m].iloc[0]
                print(f"  {t} {m}: {value:.3f}")
        # Confusion matrix
        conf_mtx = confusion_matrix(y_validate, preds, labels=label_names)
        print("\nConfusion Matrix:")
        for i, row_name in enumerate(label_names):
            for j, col_name in enumerate(label_names):
                value = conf_mtx[i, j]
                print(f"  {row_name} -> {col_name}: {value}")
    else:
        # Regression metrics
        rmse = root_mean_squared_error(y_validate, preds)
        mae = mean_absolute_error(y_validate, preds)
        r2 = r2_score(y_validate, preds)
        print("\nRegression Metrics:")
        print(f"  RMSE: {rmse:.3f}")
        print(f"  MAE: {mae:.3f}")
        print(f"  R2: {r2:.3f}")
        print(f"  NumRows: {len(df_result)}")


if __name__ == "__main__":
    # Test both classifier and regression
    print("=" * 50)
    test_model_training("classifier")

    print("\n" + "=" * 50)
    test_model_training("regression")



"""FeatureSets Layout: Layout for the FeatureSets page in the Artifact Viewer"""

from dash import html, dcc
import dash_bootstrap_components as dbc

from workbench.web_interface.components.plugins.ag_table import AGTable
from workbench.utils.theme_manager import ThemeManager
from workbench.utils.config_manager import ConfigManager

# Get the UI update rate
update_rate = ConfigManager().ui_update_rate() * 1000  # Convert to Milliseconds

# Get the Project Name
tm = ThemeManager()
project_name = tm.branding().get("project_name", "Workbench")


def feature_sets_layout(
    feature_sets_table: AGTable,
    feature_set_sample_rows: AGTable,
    feature_set_details: dcc.Markdown,
    violin_plot: dcc.Graph,
    correlation_matrix: dcc.Graph,
) -> html.Div:
    # The layout for the FeatureSets page
    layout = html.Div(
        children=[
            dcc.Interval(id="feature_sets_refresh", interval=update_rate),
            dcc.Store(id="feature_sets_page_loaded", data=False),
            dbc.Row(
                [
                    html.H2(f"{project_name}: FeatureSets"),
                ]
            ),
            # A table that lists out all the Feature Sets
            dbc.Row(feature_sets_table),
            # Sample Rows for the selected Feature Set
            dbc.Row(
                html.H3("Sampled Rows", id="feature_sample_rows_header", style={"margin": "30px 0px 10px 0px"}),
            ),
            dbc.Row(
                feature_set_sample_rows,
            ),
            # Row: Data Source Details, Column2: Violin Plots, Correlation Matrix
            dbc.Row(
                [
                    # Column 1: Feature Set Details
                    dbc.Col(
                        dcc.Loading(feature_set_details, type="dot", color="#33aa33", delay_show=300),
                        width=4,
                        className="text-break workbench-container",
                        style={"margin": "20px 0px 0px 0px", "padding": "20px"},
                    ),
                    # Column 2: Violin Plots and Correlation Matrix
                    dbc.Col(
                        [
                            dbc.Row(
                                dcc.Loading(violin_plot, type="graph", delay_show=500),
                                className="workbench-container",
                                style={"margin": "20px 0px 10px 20px"},
                            ),
                            dbc.Row(
                                dcc.Loading(correlation_matrix, type="graph", delay_show=500),
                                className="workbench-container",
                                style={"margin": "20px 0px 10px 20px"},
                            ),
                        ],
                        width=8,
                        style={"padding": "0px"},
                    ),
                ]
            ),
        ],
        style={"margin": "30px"},
    )
    return layout



"""Welcome to the Workbench DataLoaders Light Classes

These classes provide low-level APIs for loading smaller data into AWS services

- CSVToDataSource: Loads local CSV data into a DataSource
- JSONToDataSource: Loads local JSON data into a DataSource
- S3ToDataSourceLight: Loads S3 data into a DataSource
"""

from .csv_to_data_source import CSVToDataSource
from .json_to_data_source import JSONToDataSource
from .s3_to_data_source_light import S3ToDataSourceLight

__all__ = ["CSVToDataSource", "JSONToDataSource", "S3ToDataSourceLight"]



"""Layout for the Pipelines page"""

from typing import Any
from dash import html, dcc
import dash_bootstrap_components as dbc

from workbench.web_interface.components.plugins.ag_table import AGTable
from workbench.utils.config_manager import ConfigManager

# Get the UI update rate
update_rate = ConfigManager().ui_update_rate() * 1000  # Convert to Milliseconds


def pipelines_layout(
    pipelines_table: AGTable,
    pipeline_details: dcc.Markdown,
    **kwargs: Any,
) -> html.Div:
    # Generate rows for each plugin
    plugin_rows = [
        dbc.Row(
            plugin,
            style={"padding": "0px 0px 0px 0px"},
        )
        for component_id, plugin in kwargs.items()
    ]
    layout = html.Div(
        children=[
            dcc.Interval(id="pipelines_refresh", interval=update_rate),
            dcc.Store(id="pipelines_page_loaded", data=False),
            dbc.Row(
                [
                    html.H2("Workbench: Pipelines"),
                    html.Div(id="dev_null", style={"display": "none"}),  # Output for callbacks without outputs
                ]
            ),
            # A table that lists out all the Models
            dbc.Row(pipelines_table),
            # Model Details, and Plugins
            dbc.Row(
                [
                    # Column 1: Model Details
                    dbc.Col(
                        [
                            dbc.Row(
                                html.H3("Details", id="pipeline_details_header"),
                                style={"padding": "30px 0px 10px 0px"},
                            ),
                            dbc.Row(
                                pipeline_details,
                                style={"padding": "0px 0px 30px 0px"},
                            ),
                        ],
                        width=4,
                        className="text-break",
                    ),
                    # Column 2: Pipeline Metrics and Plugins
                    dbc.Col(
                        [
                            dbc.Row(
                                html.H3("Pipeline Metrics", id="pipeline_metrics_header"),
                                style={"padding": "30px 0px 0px 0px"},
                            ),
                            dbc.Row(
                                html.H3("TBD Plots", id="tbd_plots"),
                                style={"padding": "0px 0px 0px 0px"},
                            ),
                            dbc.Row(
                                html.H3("Plugins", id="plugins_header"),
                                style={"padding": "30px 0px 10px 0px"},
                            ),
                            # Add the dynamically generated Plugin rows
                            *plugin_rows,
                        ],
                        width=8,
                    ),
                ]
            ),
        ],
        style={"margin": "30px"},
    )
    return layout



"""A Markdown Component for details/information about Endpoints"""

import logging

# Dash Imports
from dash import html, dcc

# Workbench Imports
from workbench.api import Endpoint
from workbench.utils.markdown_utils import health_tag_markdown
from workbench.web_interface.components.plugin_interface import PluginInterface, PluginPage, PluginInputType

# Get the Workbench logger
log = logging.getLogger("workbench")


class EndpointDetails(PluginInterface):
    """Endpoint Details Composite Component"""

    """Initialize this Plugin Component Class with required attributes"""
    auto_load_page = PluginPage.NONE
    plugin_input_type = PluginInputType.ENDPOINT

    def __init__(self):
        """Initialize the EndpointDetails plugin class"""
        self.component_id = None
        self.current_endpoint = None

        # Call the parent class constructor
        super().__init__()

    def create_component(self, component_id: str) -> html.Div:
        """Create a Markdown Component without any data.
        Args:
            component_id (str): The ID of the web component
        Returns:
            html.Div: A Container of Components for the Model Details
        """
        self.component_id = component_id
        container = html.Div(
            id=self.component_id,
            children=[
                html.H3(id=f"{self.component_id}-header", children="Endpoint: Loading..."),
                dcc.Markdown(id=f"{self.component_id}-details"),
            ],
        )

        # Fill in plugin properties
        self.properties = [(f"{self.component_id}-header", "children"), (f"{self.component_id}-details", "children")]

        # Return the container
        return container

    def update_properties(self, endpoint: Endpoint, **kwargs) -> list:
        """Update the properties for the plugin.

        Args:
            endpoint (Endpoint): An instantiated Endpoint object
            **kwargs: Additional keyword arguments (unused)

        Returns:
            list: A list of the updated property values for the plugin
        """
        log.important(f"Updating Plugin with Endpoint: {endpoint.name} and kwargs: {kwargs}")

        # Update the header and the details
        self.current_endpoint = endpoint
        header = f"{self.current_endpoint.name}"
        details = self.endpoint_details()

        # Return the updated property values for the plugin
        return [header, details]

    def endpoint_details(self):
        """Construct the markdown string for the endpoint details

        Returns:
            str: A markdown string
        """
        # Get these fields from the endpoint
        show_fields = ["health_tags", "input", "status", "instance", "variant"]

        # Construct the markdown string
        summary = self.current_endpoint.details()
        markdown = ""
        for key in show_fields:

            # Special case for the health tags
            if key == "health_tags":
                markdown += health_tag_markdown(summary.get(key, []))
                continue

            # Get the value
            value = summary.get(key, "-")

            # If the value is a list, convert it to a comma-separated string
            if isinstance(value, list):
                value = ", ".join(value)

            # Chop off the "workbench_" prefix
            key = key.replace("workbench_", "")

            # Add to markdown string
            markdown += f"**{key}:** {value}  \n"

        return markdown


if __name__ == "__main__":
    # This class takes in endpoint details and generates a details Markdown component
    from workbench.web_interface.components.plugin_unit_test import PluginUnitTest

    # Run the Unit Test on the Plugin
    PluginUnitTest(EndpointDetails).run()



"""This Script creates the Abalone and AQSol Quantile Regression Artifacts in AWS/Workbench

FeatureSets:
    - test-quantile
Models:
    - test-quantile
    - abalone-quantile
    - aqsol-quantile
Endpoints:
    - test-quantile
    - abalone-quantile
    - aqsol-quantile
"""

import logging

from workbench.api import FeatureSet, Model, ModelType, Endpoint
from workbench.core.transforms.pandas_transforms import PandasToFeatures
from workbench.utils.plot_utils import generate_heteroskedastic_data

log = logging.getLogger("workbench")


if __name__ == "__main__":

    # Recreate Flag in case you want to recreate the artifacts
    recreate = False

    # Create the Test Quantile Regression FeatureSet
    if recreate or not FeatureSet("test_quantile").exists():
        # Generate a synthetic dataset with heteroskedasticity
        df = generate_heteroskedastic_data(n=1000)
        to_features = PandasToFeatures("test_quantile")
        to_features.set_output_tags(["test", "quantiles"])
        to_features.set_input(df, id_column="id")
        to_features.transform()

    # Create the Test Quantile Regression Model
    if recreate or not Model("test-quantile").exists():
        # Transform FeatureSet into Quantile Regression Model
        feature_set = FeatureSet("test_quantile")
        feature_set.to_model(
            name="test-quantile",
            model_type=ModelType.UQ_REGRESSOR,
            target_column="y",
            description="Test Quantile Regression",
            tags=["test", "quantiles"],
        )

    # Create the Test Quantile Regression Endpoint
    if recreate or not Endpoint("test-quantile").exists():
        m = Model("test-quantile")
        end = m.to_endpoint(tags=["test", "quantiles"])

        # Run auto-inference on the Test Quantile Regression Endpoint
        end.auto_inference(capture=True)

    # Create the Abalone Quantile Regression Model
    if recreate or not Model("abalone-quantile").exists():

        # Transform FeatureSet into Quantile Regression Model
        feature_set = FeatureSet("abalone_features")
        feature_set.to_model(
            name="abalone-quantile",
            model_type=ModelType.UQ_REGRESSOR,
            target_column="class_number_of_rings",
            description="Abalone Quantile Regression",
            tags=["abalone", "quantiles"],
        )

    # Create the Abalone Quantile Regression Endpoint
    if recreate or not Endpoint("abalone-quantile").exists():
        m = Model("abalone-quantile")
        end = m.to_endpoint(tags=["abalone", "quantiles"])

        # Run auto-inference on the Abalone Quantile Regression Endpoint
        end.auto_inference(capture=True)

    # Create the AQSol Quantile Regression Model
    if recreate or not Model("aqsol-quantiles").exists():

        # AQSol Features
        features = [
            "molwt",
            "mollogp",
            "molmr",
            "heavyatomcount",
            "numhacceptors",
            "numhdonors",
            "numheteroatoms",
            "numrotatablebonds",
            "numvalenceelectrons",
            "numaromaticrings",
            "numsaturatedrings",
            "numaliphaticrings",
            "ringcount",
            "tpsa",
            "labuteasa",
            "balabanj",
            "bertzct",
        ]

        # Transform FeatureSet into Quantile Regression Model
        feature_set = FeatureSet("aqsol_features")
        feature_set.to_model(
            name="aqsol-quantiles",
            model_type=ModelType.UQ_REGRESSOR,
            target_column="solubility",
            feature_list=features,
            description="AQSol Quantile Regression",
            tags=["aqsol", "quantiles"],
        )

    # Create the AQSol Quantile Regression Endpoint
    if recreate or not Endpoint("aqsol-quantiles").exists():
        m = Model("aqsol-quantiles")
        end = m.to_endpoint(tags=["aqsol", "quantiles"])

        # Run auto-inference on the AQSol Quantile Regression Endpoint
        end.auto_inference(capture=True)



from dash import dcc, html, callback, Input, Output
import plotly.graph_objects as go
from dash.exceptions import PreventUpdate
import networkx as nx

# Workbench Imports
from workbench.web_interface.components.plugin_interface import PluginInterface, PluginPage, PluginInputType
from workbench.utils.theme_manager import ThemeManager
from workbench.utils.color_utils import weights_to_colors


class GraphPlot(PluginInterface):
    """A Graph Plot Plugin for NetworkX Graphs."""

    # Initialize this Plugin Component Class with required attributes
    auto_load_page = PluginPage.NONE
    plugin_input_type = PluginInputType.GRAPH

    def __init__(self):
        """Initialize the Graph (Nodes/Edges) Plugin"""
        self.component_id = None
        self.hover_columns = []
        self.graph = None
        self.graph_figure = None

        # Initialize the Theme Manager
        self.theme_manager = ThemeManager()

        # Call the parent class constructor
        super().__init__()

    def create_component(self, component_id: str) -> html.Div:
        """Create a Graph (Node/Edge) Component without any data.

        Args:
            component_id (str): The ID of the web component.

        Returns:
            html.Div: A Container of Components for the Graph (Node/Edge) Component.
        """
        self.component_id = component_id

        # Fill in plugin properties and signals
        self.properties = [
            (f"{component_id}-graph", "figure"),
            (f"{component_id}-label-dropdown", "options"),
            (f"{component_id}-color-dropdown", "options"),
            (f"{component_id}-label-dropdown", "value"),
            (f"{component_id}-color-dropdown", "value"),
        ]
        self.signals = [(f"{component_id}-graph", "hoverData")]

        # Create the Composite Component
        # - A Graph Node/Edge Component
        # - Dropdown for Node Labels and Colors
        return html.Div(
            children=[
                # Main Scatter Plot Graph
                dcc.Graph(
                    id=f"{component_id}-graph",
                    figure=self.display_text("Waiting for Data..."),
                    config={"scrollZoom": True},
                    style={"width": "100%", "height": "100%"},  # Let the graph fill its container
                ),
                # Controls: Label, Color Dropdowns, and TBD Checkbox
                html.Div(
                    [
                        html.Label("Label", style={"marginLeft": "40px", "marginRight": "5px", "fontWeight": "bold"}),
                        dcc.Dropdown(
                            id=f"{component_id}-label-dropdown",
                            className="dropdown",
                            style={"min-width": "50px", "flex": 1},  # Responsive width
                            clearable=False,
                        ),
                        html.Label("Color", style={"marginLeft": "30px", "marginRight": "5px", "fontWeight": "bold"}),
                        dcc.Dropdown(
                            id=f"{component_id}-color-dropdown",
                            className="dropdown",
                            style={"min-width": "50px", "flex": 1},  # Responsive width
                            clearable=False,
                        ),
                        dcc.Checklist(
                            id=f"{component_id}-tbd-checkbox",
                            options=[{"label": " TBD", "value": "show"}],
                            value=[],
                            style={"margin": "10px"},
                        ),
                    ],
                    style={"padding": "10px", "display": "flex", "gap": "10px"},
                ),
            ],
            style={"height": "100%", "display": "flex", "flexDirection": "column"},  # Full viewport height
        )

    def update_properties(self, input_graph: nx.Graph, **kwargs) -> list:
        """Update the property values for the plugin component.

        Args:
            input_graph (NetworkX Graph): The input graph data object.
            **kwargs: Additional keyword arguments (plugins can define their own arguments).
                      Note: The current kwargs processed are:
                            - label: The default node labels
                            - color: The default color column
                            - label_columns: The columns to use for the labels
                            - color_columns: The columns to use for the color scale
                            - hover_columns: The columns to show when hovering over a node

        Returns:
            list: A list of updated properties (figure, label_list, color_list, default_label, default)color).
        """

        # Get the NetworkX graph
        self.graph = input_graph

        # We'll use the first node to look for node attributes
        first_node = self.graph.nodes[next(iter(self.graph.nodes))]

        # Check if nodes have 'x' and 'y' attributes
        if "x" not in first_node or "y" not in first_node:
            self.log.important("No 'x'/'y' attributes found, running layout...")

            # Choose layout algorithm based on graph size
            node_count = len(self.graph.nodes())
            if node_count < 500:
                self.log.important(f"Graph has {node_count} nodes, using spring layout")
                pos = nx.spring_layout(self.graph, iterations=500)
            else:
                self.log.important(f"Large graph with {node_count} nodes, using spectral layout")
                pos = nx.spectral_layout(self.graph)

            # Update x/y attributes for all nodes
            for node, position in pos.items():
                self.graph.nodes[node]["x"] = position[0]
                self.graph.nodes[node]["y"] = position[1]

        # Use 'id' as default label field if not specified
        label_field = kwargs.get("label", next(iter(first_node), "id"))

        # Add degree attribute if not already present in nodes
        nx.set_node_attributes(self.graph, dict(self.graph.degree()), "degree")

        # Extract positions, labels, degrees, and hover text in a single pass through the nodes
        x_nodes, y_nodes, labels, node_degrees, hover_text = [], [], [], [], []

        # Define hover columns if not specified
        hover_columns = kwargs.get("hover_columns", ["id", "degree"])  # Default hover columns
        if hover_columns == "all":
            hover_columns = [key for key in first_node if key != "x" and key != "y"]

        # Iterate through nodes once and extract required fields
        for node, data in self.graph.nodes(data=True):
            x_nodes.append(data["x"])
            y_nodes.append(data["y"])
            labels.append(data.get(label_field, ""))
            node_degrees.append(data.get("degree", self.graph.degree[node]))
            hover_text.append("<br>".join([f"{key}: {data.get(key, '')}" for key in hover_columns]))

        # Node label colors
        if self.theme_manager.dark_mode():
            label_color = "rgba(200, 200, 255)"
        else:
            label_color = "rgba(100, 100, 100)"

        # Create an OpenGL Scattergl plot for nodes using Plotly
        node_trace = go.Scattergl(
            x=x_nodes,
            y=y_nodes,
            mode="markers+text",  # Include text mode for labels
            text=labels,  # Set text for node labels
            textfont=dict(color=label_color, size=14),  # Set font size/color for node labels
            textposition="top center",  # Position labels above the nodes
            hovertext=hover_text,  # Set hover text for nodes
            hovertemplate="%{hovertext}<extra></extra>",  # Define hover template and remove extra info
            marker=dict(
                size=20,  # Marker size for nodes
                color=node_degrees,  # Use node degrees for marker colors
                colorbar=dict(title="Degree"),  # Include a color bar for degrees
                line=dict(color="rgba(0, 0, 0, 0.5)", width=2),  # Set border color and width for nodes
            ),
        )

        # Create Scattergl traces for edges in a single loop for efficiency
        edge_traces = []
        color_scale = self.theme_manager.colorscale()
        edge_width = 5
        edge_weights = list(nx.get_edge_attributes(self.graph, "weight").values())
        edge_colors = weights_to_colors(edge_weights, color_scale)
        for edge, color in zip(self.graph.edges(), edge_colors):
            x0, y0 = self.graph.nodes[edge[0]]["x"], self.graph.nodes[edge[0]]["y"]
            x1, y1 = self.graph.nodes[edge[1]]["x"], self.graph.nodes[edge[1]]["y"]

            # Create individual Scattergl trace for each edge with specific styling
            edge_traces += [
                go.Scattergl(
                    x=[x0, x1],
                    y=[y0, y1],
                    mode="lines",
                    line=dict(width=edge_width + 2, color="rgba(0, 0, 0, 0.5)"),  # Edge drop shadow
                    showlegend=False,
                    hoverinfo="skip",  # Skip hover info for edges if not needed
                ),
                go.Scattergl(
                    x=[x0, x1],
                    y=[y0, y1],
                    mode="lines",
                    line=dict(width=edge_width, color=color),
                    showlegend=False,
                    hoverinfo="skip",  # Skip hover info for edges if not needed
                ),
            ]

        # Create a Plotly figure with the combined node and edge traces
        self.graph_figure = go.Figure(data=edge_traces + [node_trace])

        # Fine-tune the plot's layout and aesthetics
        self.graph_figure.update_layout(
            margin={"t": 10, "b": 10, "r": 10, "l": 10, "pad": 10},  # Set margins and padding
            xaxis=dict(showgrid=False, showticklabels=False, zeroline=False),  # Hide X-axis grid and tick marks
            yaxis=dict(showgrid=False, showticklabels=False, zeroline=False),  # Hide Y-axis grid and tick marks
            showlegend=False,  # Remove legend
            dragmode="pan",
            modebar={"bgcolor": "rgba(0, 0, 0, 0)"},  # Transparent background for modebar
            uirevision="constant",  # Keep the layout constant for updates
        )

        # Get the first node's attributes (all fields should be defined here)
        node_attributes = {key for key in first_node.keys()}
        node_attributes.discard("x")  # Remove 'x/y' from the attributes
        node_attributes.discard("y")

        # Create dropdown options for the label list (all non-'x/y' attributes)
        label_list = [{"label": "None", "value": "None"}]  # Add "None" option as the first entry
        label_list.extend([{"label": attr, "value": attr} for attr in node_attributes])

        # Create the color list using only numeric attributes or enumerated string fields
        color_list = []
        for attr in node_attributes:
            sample_value = first_node[attr]

            # Check if the attribute is numeric
            if isinstance(sample_value, (int, float)):
                color_list.append({"label": attr, "value": attr})
            # If it's a string, add the option with a note about enumeration
            elif isinstance(sample_value, str):
                color_list.append({"label": f"{attr} (enum)", "value": attr})

        # Set default dropdown values for label and color
        default_label = label_field  # Use the specified label field as default
        default_color = "degree" if "degree" in node_attributes else next(iter(node_attributes), "id")

        # Return the updated properties for the dropdowns and the figure
        return [self.graph_figure, label_list, color_list, default_label, default_color]

    def register_internal_callbacks(self):
        """Register any internal callbacks for the plugin."""

        @callback(
            Output(f"{self.component_id}-graph", "figure", allow_duplicate=True),
            [
                Input(f"{self.component_id}-label-dropdown", "value"),
                Input(f"{self.component_id}-color-dropdown", "value"),
            ],
            prevent_initial_call=True,
        )
        def update_graph(label, color):
            """Update the Graph (Nodes/Edges) based on the dropdown values."""
            if not label and not color:
                raise PreventUpdate

            # Use the class variable to access the graph's nodes directly
            nodes = self.graph.nodes

            # Get the node trace from the graph figure (assuming last trace is for nodes)
            node_trace = self.graph_figure["data"][-1]

            # Update node labels dynamically if a label field is selected
            if label:
                if label == "None":  # Handle "None" option
                    node_trace["text"] = [""] * len(nodes)
                else:
                    node_trace["text"] = [nodes[node].get(label, "") for node in nodes]

            # Update node colors dynamically if a color field is selected
            if color:
                # Check if the attribute exists and if it's numeric
                first_node = next(iter(nodes))
                if color in nodes[first_node]:
                    color_values = [nodes[node][color] for node in nodes]
                    if isinstance(nodes[first_node][color], (int, float)):
                        node_trace["marker"]["color"] = color_values
                    else:
                        # Enumerate strings as a fallback for categorical values
                        unique_values = {val: idx for idx, val in enumerate(set(color_values))}
                        node_trace["marker"]["color"] = [unique_values[val] for val in color_values]
                else:
                    node_trace["marker"]["color"] = [0] * len(nodes)  # Default to zero if not present

            # Return the updated figure
            return self.graph_figure


if __name__ == "__main__":
    """Run the Unit Test for the Plugin."""
    from workbench.web_interface.components.plugin_unit_test import PluginUnitTest
    from workbench.api.graph_store import GraphStore

    # Pull a test graph
    graph_store = GraphStore()
    test_graph = graph_store.get("chem_info/tox21_100")
    if test_graph is None:
        print("Test graph not found... using default")

    # Run the Unit Test on the Plugin
    PluginUnitTest(GraphPlot, theme="dark", input_data=test_graph).run()



import os
import re
import logging
from openai import OpenAI

from workbench.utils.workbench_cache import WorkbenchCache


class AISummary:
    def __init__(self):
        """
        Initialize the AISummary class by setting up the OpenAI client.
        """
        self.log = logging.getLogger("workbench")

        # Get the API key from the environment
        self.deepseek_api = os.getenv("DEEPSEEK_API")

        # If the API key is not set, raise an error
        if not self.deepseek_api:
            raise ValueError("DEEPSEEK API key is not set")

        # Create a client with the API key
        self.client = OpenAI(api_key=self.deepseek_api, base_url="https://api.deepseek.com/v1")
        self.ai_cache = WorkbenchCache(prefix="ai_summary")

    def smiles_query(self, smiles_string: str, force_pull=False) -> str:
        """
        Query the DeepSeek API for information about a compound using its SMILES string.

        Args:
            smiles_string (str): The SMILES string of the compound.
            force_pull (bool, optional): Force the API to pull the data again. Defaults to False.

        Returns:
            str: A markdown-formatted bulleted list of the compound's properties.
        """
        # Check if the summary is already cached
        cached_summary = self.ai_cache.get(smiles_string)
        if cached_summary and not force_pull:
            self.log.info(f"Using cached summary for SMILES: {smiles_string}")
            return cached_summary

        # If the summary is not cached, query the API
        self.log.info(f"Querying the API for SMILES: {smiles_string}")

        # Define the task and lookup context
        task = """
            Search open sources and use internal chemistry knowledge to provide a concise
            summary of its properties relevant to targeted therapeutics and toxicity.
            DO NOT INCLUDE INTRODUCTORY SENTENCES, CONCLUSIONS, OR CONVERSATIONAL PHRASES
        """
        lookup_context = {
            "action": "lookup",
            "resources": ["PubMed", "PubChem", "ChEMBL", "DrugBank", "ChemSpider", "open source"],
            "query": f"Find information about the compound with SMILES {smiles_string}. {task}",
        }

        try:
            # Make the API request
            response = self.client.chat.completions.create(
                model="deepseek-chat",  # Specify the model
                messages=[{"role": "user", "content": str(lookup_context)}],  # Convert dict to string
            )

            # Extract the summary from the response
            summary = response.choices[0].message.content

            # If one of the lines has SMILES, delete that line
            summary = "\n".join([line for line in summary.split("\n") if "SMILES" not in line])

            # Clean up the summary
            # summary = self.clean_summary(summary)

            # Cache the summary for future use
            self.ai_cache.set(smiles_string, summary)
            return summary

        except Exception as e:
            # Handle any errors that occur during the API request
            return f"### Error\n\nAn error occurred while querying the API: {str(e)}"

    @staticmethod
    def clean_summary(response: str) -> str:
        # Remove leading numbers with dots (e.g., "1. ", "2. ")
        response = re.sub(r"^\d+\.\s", "", response, flags=re.MULTILINE)

        # Strip whitespace from each line
        cleaned_lines = [line.strip() for line in response.split("\n")]

        # Join the cleaned lines into a single string
        cleaned_response = "\n".join(cleaned_lines)

        return cleaned_response


# Example usage
if __name__ == "__main__":
    # Example SMILES from tox21 dataset

    smiles_string = "CC(C)C1=CC=C(C=C1)C(=O)O"
    smiles_string = "CCCC[n+]1cccc(C)c1.F[B-](F)(F)F"
    # smiles_string = "CCCCCCCCCCCCCC[n+]1ccc(C)cc1.[Cl-]"

    # Create an instance of the AISummary class
    ai_summary = AISummary()

    # Query the API and get the markdown summary
    summary_markdown = ai_summary.smiles_query(smiles_string, force_pull=True)

    # Print the markdown result
    print(summary_markdown)



from workbench.api import Model, ModelType
from workbench.core.transforms.features_to_model.features_to_model import FeaturesToModel
from workbench.utils.model_utils import get_custom_script_path

fs_name = "solubility_featurized_class_0_fs"
fs_name = "solubility_test_features"
# fs_name = "aqsol_features"


# Molecular Descriptors Custom Model
my_script_path = get_custom_script_path("chem_info", "molecular_descriptors.py")

to_model = FeaturesToModel(
    fs_name, "smiles-to-taut-md-stereo-v0", model_type=ModelType.TRANSFORMER, custom_script=my_script_path
)
to_model.set_output_tags(["smiles", "molecular descriptors"])
to_model.transform(target_column=None, feature_list=["smiles"], description="Smiles to Molecular Descriptors")

# Deploy an Endpoint for the Model
my_model = Model("smiles-to-taut-md-stereo-v0")
endpoint = my_model.to_endpoint(name="smiles-to-taut-md-stereo-v0", tags=["smiles", "molecular descriptors"])

# Run auto-inference on the Endpoint
endpoint.auto_inference(capture=True)



import redis


def delete_keys_by_pattern(redis_host, redis_port, pattern):
    # Connect to the Redis instance
    r = redis.Redis(host=redis_host, port=redis_port, decode_responses=True)

    # Use the SCAN command to find all keys matching the pattern
    cursor = 0
    while True:
        cursor, keys = r.scan(cursor, match=pattern, count=1000)
        for key in keys:
            print(f"Deleting {key}...")
            r.unlink(key)
        if cursor == 0:
            break


if __name__ == "__main__":
    redis_host = "workbenchredis.qo8vb5.0001.use1.cache.amazonaws.com"
    redis_port = 6379
    pattern = "workbench:RowStorage:*"

    delete_keys_by_pattern(redis_host, redis_port, pattern)



"""Tests for graph_plot web component"""

import time

# Workbench Imports
from workbench.api.graph_store import GraphStore
from workbench.web_interface.components.plugins.graph_plot import GraphPlot
from workbench.utils.theme_manager import ThemeManager


def test_graph_plot():
    """Test the GraphPlot class"""

    # Set the theme
    ThemeManager().set_theme("light")

    # Instantiate a Graph
    graph = GraphStore().get("test/karate_club")

    # Instantiate the GraphPlot class
    graph_plot = GraphPlot()

    # Generate the figure
    properties = graph_plot.update_properties(graph, label="club")

    # The first property should be the figure
    figure = properties[0]
    figure.show()

    # Sleep for a second (for plot to show)
    time.sleep(1)


if __name__ == "__main__":
    # Run the tests
    test_graph_plot()



from rdkit import Chem
from workbench.api.df_store import DFStore
from workbench.utils.chem_utils import toxic_elements, toxic_groups

# Grab the tox21 data
tox21 = DFStore().get("/datasets/chem_info/tox21")

# Add molecules to the dataframe
tox21["molecule"] = tox21["smiles"].apply(lambda smiles: Chem.MolFromSmiles(smiles))

# See which molecules are tagged as toxic but not toxic (based on the tox21 dataset)
too_broad = tox21[(tox21["toxic_tag"] == 1) & (tox21["toxic_any"] == 0)].copy()

# Let's check to see if the too_broad molecules still get flagged as toxic
too_broad["toxic_elements"] = too_broad["molecule"].apply(toxic_elements)
too_broad["toxic_groups"] = too_broad["molecule"].apply(toxic_groups)


# Define toxic SMARTS patterns
smarts_patterns = [
    "[N+](=O)[O-]",  # Nitro group (corrected)
    "[C](Cl)(Cl)(Cl)",  # Trichloromethyl group
    "C(=S)N",  # Dithiocarbamate group
    "[Cr](=O)(=O)=O",  # Chromium(VI)
    "[Se][Se]",  # Diselenide group
    "S(=O)(=O)O",  # Sulfonates
    "[N+](C)(C)CCOc1ccccc1",  # Quaternary ammonium with aromatic ether
    "[N+](C)(C)(C)",  # Quaternary ammonium
    "[N+](C)(C)C",  # Tertiary ammonium
    "[n+](N)no",  # Nitro heterocycle with amine
    "[CBr]",  # Organobromides with high halogenation
    "[n+]1cccc(CO)c1",  # Nitroso compound with CO group
]

# Apply Chem.MolFromSmarts() to each pattern
toxic_smarts = [Chem.MolFromSmarts(smarts) for smarts in smarts_patterns]


# Function to apply SMARTS filtering
def filter_known_toxic(smiles, toxic_smarts):
    mol = Chem.MolFromSmiles(smiles)
    if not mol:
        return False  # Skip invalid SMILES
    return any(mol.HasSubstructMatch(smarts) for smarts in toxic_smarts)


# Apply filter
# print(f"False Positives: {len(too_broad)}")
# too_broad = too_broad[~too_broad["smiles"].apply(lambda x: filter_known_toxic(x, toxic_smarts))]
# print(f"False Positives after filtering: {len(too_broad)}")

# Print out the results
print(too_broad[["toxic_any", "toxic_elements", "toxic_groups"]].value_counts(dropna=False))

# Grab the ones marked with toxic elements
my_toxic_elements = too_broad[too_broad["toxic_elements"]].copy()

# Print out the results
print(f"False Positives Elements: {len(my_toxic_elements)}")
print(my_toxic_elements["smiles"].values[:10].tolist())

# Grab the ones marked with toxic groups
my_toxic_groups = too_broad[too_broad["toxic_groups"]].copy()

# Print out the results
print(f"False Positives Groups: {len(my_toxic_groups)}")
print(my_toxic_groups["smiles"].values[:10].tolist())

# Too narrow: See which molecules are not tagged as toxic but ARE toxic (based on the tox21 dataset)
too_narrow = tox21[(tox21["toxic_tag"] == 0) & (tox21["toxic_any"] == 1)].copy()

# Let's check to see if the too_narrow molecules are still flagged as not toxic
too_narrow["toxic_elements"] = too_narrow["molecule"].apply(toxic_elements)
too_narrow["toxic_groups"] = too_narrow["molecule"].apply(toxic_groups)


# Print out the results
print(too_narrow[["toxic_any", "toxic_elements", "toxic_groups"]].value_counts(dropna=False))

# Grab the ones not flagged by either toxic elements or toxic groups (so two false negatives)
my_toxic_elements = too_narrow[~too_narrow["toxic_elements"] & ~too_narrow["toxic_groups"]].copy()

# Print out the results
print(f"False Negatives: {len(my_toxic_elements)}")
print(list(my_toxic_elements["smiles"].values[:10].tolist()))



from sagemaker.pipeline import PipelineModel

# Workbench imports
from workbench.api import Model
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp

session = AWSAccountClamp().sagemaker_session()
role = AWSAccountClamp().aws_session.get_workbench_execution_role_arn()

# Our three models
taut_model = Model("tautomerize-v0").sagemaker_model_object()
md_model = Model("smiles-to-taut-md-stereo-v0").sagemaker_model_object()
sol_model = Model("aqsol-mol-class").sagemaker_model_object()

# Create a pipeline model that chains the three models
pipeline_model = PipelineModel(
    name="pipeline-model", models=[taut_model, md_model, sol_model], role=role, sagemaker_session=session
)

# Deploy the pipeline endpoint
predictor = pipeline_model.deploy(initial_instance_count=1, instance_type="ml.m5.large")



"""CachedFeatureSet: Caches the method results for Workbench FeatureSets"""

from typing import Union

import pandas as pd

# Workbench Imports
from workbench.core.artifacts.feature_set_core import FeatureSetCore
from workbench.core.artifacts.cached_artifact_mixin import CachedArtifactMixin


class CachedFeatureSet(CachedArtifactMixin, FeatureSetCore):
    """CachedFeatureSet: Caches the method results for Workbench FeatureSets

    Note: Cached method values may lag underlying FeatureSet changes.

    Common Usage:
        ```python
        my_features = CachedFeatureSet(name)
        my_features.details()
        my_features.health_check()
        my_features.workbench_meta()
        ```
    """

    def __init__(self, feature_set_name: str, database: str = "workbench"):
        """CachedFeatureSet Initialization"""
        FeatureSetCore.__init__(self, feature_set_name=feature_set_name, use_cached_meta=True)

    @CachedArtifactMixin.cache_result
    def summary(self, **kwargs) -> dict:
        """Retrieve the FeatureSet Details.

        Returns:
            dict: A dictionary of details about the FeatureSet
        """
        return super().summary(**kwargs)

    @CachedArtifactMixin.cache_result
    def details(self, **kwargs) -> dict:
        """Retrieve the FeatureSet Details.

        Returns:
            dict: A dictionary of details about the FeatureSet
        """
        return super().details(**kwargs)

    @CachedArtifactMixin.cache_result
    def health_check(self, **kwargs) -> dict:
        """Retrieve the FeatureSet Health Check.

        Returns:
            dict: A dictionary of health check details for the FeatureSet
        """
        return super().health_check(**kwargs)

    @CachedArtifactMixin.cache_result
    def workbench_meta(self) -> Union[str, None]:
        """Retrieve the Workbench Metadata for this DataSource.

        Returns:
            Union[dict, None]: Dictionary of Workbench metadata for this Artifact
        """
        return super().workbench_meta()

    @CachedArtifactMixin.cache_result
    def smart_sample(self) -> pd.DataFrame:
        """Retrieve the Smart Sample for this FeatureSet.

        Returns:
            pd.DataFrame: The Smart Sample DataFrame
        """
        return super().smart_sample()


if __name__ == "__main__":
    """Exercise the CachedFeatureSet Class"""
    from pprint import pprint

    # Retrieve an existing FeatureSet
    my_features = CachedFeatureSet("abalone_features")
    pprint(my_features.smart_sample())
    pprint(my_features.summary())
    pprint(my_features.details())
    pprint(my_features.health_check())
    pprint(my_features.workbench_meta())

    # Shutdown the ThreadPoolExecutor (note: users should NOT call this)
    my_features._shutdown()



"""SQL based Descriptive Stats: Compute Descriptive Stats for all the numeric columns in a DataSource using SQL"""

import logging
import pandas as pd
from collections import defaultdict

# Workbench Imports
from workbench.core.artifacts.data_source_abstract import DataSourceAbstract


# Workbench Logger
log = logging.getLogger("workbench")


def descriptive_stats_query(columns: list[str], table_name: str) -> str:
    """Build a query to compute the descriptive stats for all columns in a table
    Args:
        columns(list(str)): The columns to compute descriptive stats on
        table_name(str): The table to compute descriptive stats on
    Returns:
        str: The SQL query to compute descriptive stats
    """
    query = f'SELECT <<column_descriptive_stats>> FROM "{table_name}"'
    column_descriptive_stats = ""
    for c in columns:
        column_descriptive_stats += (
            f'min("{c}") AS "{c}___min", '
            f'approx_percentile("{c}", 0.25) AS "{c}___q1", '
            f'approx_percentile("{c}", 0.5) AS "{c}___median", '
            f'approx_percentile("{c}", 0.75) AS "{c}___q3", '
            f'max("{c}") AS "{c}___max", '
            f'avg("{c}") AS "{c}___mean", '
            f'stddev("{c}") AS "{c}___stddev", '
        )
    query = query.replace("<<column_descriptive_stats>>", column_descriptive_stats[:-2])

    # Return the query
    return query


def descriptive_stats(data_source: DataSourceAbstract) -> dict[dict]:
    """Compute Descriptive Stats for all the numeric columns in a DataSource
    Args:
        data_source(DataSource): The DataSource that we're computing descriptive stats on
    Returns:
        dict(dict): A dictionary of descriptive stats for each column in this format
             {'col1': {'min': 0, 'q1': 1, 'median': 2, 'q3': 3, 'max': 4, 'mean': 2.5, 'stddev': 1.5},
              'col2': ...}
    """
    # Grab the DataSource computation view table name
    table = data_source.view("computation").table

    # Figure out which columns are numeric
    num_type = ["double", "float", "int", "bigint", "smallint", "tinyint"]
    details = data_source.view("computation").column_details()
    numeric = [column for column, data_type in details.items() if data_type in num_type]

    # Sanity Check for numeric columns
    if len(numeric) == 0:
        log.warning("No numeric columns found in the current computation view of the DataSource")
        log.warning("If the data source was created from a DataFrame, ensure that the DataFrame was properly typed")
        log.warning("Recommendation: Properly type the DataFrame and recreate the Workbench artifact")
        return {}

    # Build the query
    query = descriptive_stats_query(numeric, table)

    # Run the query
    log.debug(query)
    result_df = data_source.query(query)

    # Process the results
    # Note: The result_df is a DataFrame with a single row and a column for each stat metric
    stats_dict = result_df.to_dict(orient="index")[0]

    # Convert the dictionary to a nested dictionary
    # Note: The keys are in the format col1__col2
    nested_descriptive_stats = defaultdict(dict)
    for key, value in stats_dict.items():
        col1, col2 = key.split("___")
        nested_descriptive_stats[col1][col2] = value

    # Return the nested dictionary
    return dict(nested_descriptive_stats)


if __name__ == "__main__":
    """Exercise the SQL Descriptive Stats Functionality"""
    from pprint import pprint
    from workbench.api.data_source import DataSource

    # Setup Pandas output options
    pd.set_option("display.max_colwidth", 50)
    pd.set_option("display.max_columns", 15)
    pd.set_option("display.width", 1000)

    # Retrieve a Data Source
    my_data = DataSource("aqsol_data")

    # Verify that the Athena Data Source exists
    assert my_data.exists()

    # What's my Workbench Name
    print(f"Name: {my_data.name}")

    # Get descriptive_stats for numeric columns
    my_descriptive_stats = descriptive_stats(my_data)
    print("\nDescriptive Stats")
    pprint(my_descriptive_stats)



import logging
from concurrent.futures import ProcessPoolExecutor, as_completed
from workbench.api import DataSource, FeatureSet, Model, ModelType
from workbench.api import Meta
from workbench.utils.test_data_generator import TestDataGenerator

parallel_jobs = 16

# Define expected artifacts
expected_artifacts = {
    "feature_sets": [f"stress_features_{i}" for i in range(parallel_jobs)],
    "models": [f"stress-model-{i}" for i in range(parallel_jobs)],
    "endpoints": [f"stress-end-{i}" for i in range(parallel_jobs)],
}

# Set up logging
log = logging.getLogger("workbench")
log.setLevel(logging.DEBUG)


# Function to create a model pipeline for a given model name
def create_model_pipeline(model_name):
    print(f"Creating pipeline for model: {model_name}")

    # Create the test_data DataSource if it doesn't already exist
    if not DataSource("test_data").exists():
        test_data = TestDataGenerator()
        df = test_data.person_data()
        DataSource(df, name="test_data")

    # Convert model name to a FeatureSet name
    feature_set_name = model_name.replace("-model-", "_features_")

    # Create the FeatureSet
    ds = DataSource("test_data")
    ds.to_features(feature_set_name, id_column="id", event_time_column="date")

    # Create the Model
    fs = FeatureSet(feature_set_name)
    m = fs.to_model(
        name=model_name,
        model_type=ModelType.REGRESSOR,
        target_column="iq_score",
        tags=["test"],
        description=f"Test Model {model_name}",
    )
    m.set_owner("test")

    # Create the Endpoint for the Model
    endpoint_name = model_name.replace("model", "end")
    print(f"Creating endpoint {endpoint_name} for model {model_name}")
    m = Model(model_name)
    end = m.to_endpoint(endpoint_name, tags=["test"])

    # Run inference on the endpoint
    end.auto_inference(capture=True)


if __name__ == "__main__":
    """A stress test to create AWS ML pipelines in parallel."""

    # List of model names to run in parallel
    model_names = expected_artifacts["models"]

    # Use ProcessPoolExecutor to run in parallel
    with ProcessPoolExecutor(max_workers=parallel_jobs) as executor:
        futures = {executor.submit(create_model_pipeline, model_name): model_name for model_name in model_names}

        for future in as_completed(futures):
            model_name = futures[future]
            try:
                future.result()  # Retrieve result to raise any exception if it occurred
            except Exception as e:
                print(f"Exception for model '{model_name}': {e}")

    # Grab all currently deployed artifacts
    meta = Meta()
    feature_sets = meta.feature_sets()["Feature Group"].tolist()
    models = meta.models()["Model Group"].tolist()
    endpoints = meta.endpoints()["Name"].tolist()

    # Check for missing artifacts
    missing_artifacts = {
        "feature_sets": [fs for fs in expected_artifacts["feature_sets"] if fs not in feature_sets],
        "models": [m for m in expected_artifacts["models"] if m not in models],
        "endpoints": [e for e in expected_artifacts["endpoints"] if e not in endpoints],
    }

    # Display missing artifacts
    if any(missing_artifacts.values()):
        print("Missing artifacts:", missing_artifacts)
    else:
        print("All expected artifacts created successfully.")



import pytest
import workbench  # noqa: F401
import logging
from workbench.utils.test_data_generator import TestDataGenerator
from workbench.api import DataSource, FeatureSet, Model, Endpoint, ModelType

# Set the logging level
logging.getLogger("workbench").setLevel(logging.DEBUG)


def create_data_source():
    test_data = TestDataGenerator()
    df = test_data.person_data()
    if not DataSource("abc_test").ready():
        DataSource(df, name="abc_test")


def create_feature_set():
    create_data_source()

    # If the feature set doesn't exist, create it
    if not FeatureSet("abc_features").ready():
        DataSource("abc_test").to_features("abc_features", id_column="id")


def create_model():
    create_feature_set()

    # If the model doesn't exist, create it
    if not Model("abc-regression").ready():
        FeatureSet("abc_features").to_model(
            name="abc-regression", model_type=ModelType.REGRESSOR, target_column="iq_score"
        )


def create_endpoint():
    create_model()

    # Create some new endpoints
    if not Endpoint("abc-regression").ready():
        Model("abc-regression").to_endpoint(name="abc-regression")


@pytest.mark.long
def test_endpoint_deletion():
    create_endpoint()

    # Now Delete the endpoint
    Endpoint("abc-regression").delete()


@pytest.mark.long
def test_model_deletion():
    create_model()

    # Now Delete the Model
    Model("abc-regression").delete()


@pytest.mark.long
def test_feature_set_deletion():
    create_feature_set()

    # Now Delete the FeatureSet
    FeatureSet("abc_features").delete()


@pytest.mark.long
def test_data_source_deletion():
    create_data_source()

    # Now Delete the DataSource
    DataSource("abc_test").delete()


if __name__ == "__main__":

    test_endpoint_deletion()
    test_model_deletion()
    test_feature_set_deletion()
    test_data_source_deletion()



from workbench.api import DFStore
from workbench.utils.pandas_utils import compare_dataframes

# Initialize DFStore and load dataframes
df_store = DFStore()
old_df = df_store.get("/smiles_testing/hlh/old_endpoint")
new_df = df_store.get("/smiles_testing/hlh/new_endpoint")

# Convert all columns names to lowercase
old_df.columns = old_df.columns.str.lower()

# We know that the old endpoint crashes on these two smiles (so the endpoint harness puts in NaNs)
old_error_smiles_1 = "CCC1C2=N3C(=CC4=C(C(C)=O)C(C)=C5C=C6C(C)C(CCC(=O)O)C7=N6[Pd+2]3([N-]54)[N-]3C(=C2)C(C)=C(C(=O)NCCS(=O)(=O)O)C3=C7CC(=O)OC)C1C"
old_error_smiles_2 = "O=C(CCC1C(=O)[O-][Gd+3]23456[O-]C(=O)C(CCC(=O)NCC(O)CO)N27CCN13CCN4(CC1=N5C(=CC=C1)C7)C(CCC(=O)NCC(O)CO)C(=O)[O-]6)NCC(O)CO"

# Remove this smiles from both dataframes
old_df = old_df[~old_df["smiles"].isin([old_error_smiles_1, old_error_smiles_2])]
new_df = new_df[~new_df["smiles"].isin([old_error_smiles_1, old_error_smiles_2])]

# Compare the dataframes
comparison = compare_dataframes(old_df, new_df, ["udm_mol_bat_id", "smiles"])



"""Resource utilities for Workbench"""

import sys
import importlib.resources as resources
import pathlib
import pkg_resources


def get_resource_path(package: str, resource: str) -> pathlib.Path:
    """Get the path to a resource file, compatible with Python 3.9 and higher.

    Args:
        package (str): The package where the resource is located.
        resource (str): The name of the resource file.

    Returns:
        pathlib.Path: The path to the resource file.
    """
    if sys.version_info >= (3, 10):
        # Python 3.10 and higher: use importlib.resources.path
        with resources.path(package, resource) as path:
            return path
    else:
        # Python 3.9 and lower: manually construct the path based on package location
        # Get the location of the installed package
        package_location = pathlib.Path(pkg_resources.get_distribution(package.split(".")[0]).location)
        resource_path = package_location / package.replace(".", "/") / resource

        if resource_path.exists():
            return resource_path
        else:
            raise FileNotFoundError(f"Resource '{resource}' not found in package '{package}'.")


if __name__ == "__main__":
    # Test the resource utilities
    with get_resource_path("workbench.resources", "open_source_api.key") as open_source_key_path:
        with open(open_source_key_path, "r") as key_file:
            print(key_file.read().strip())



import pandas as pd
from sklearn.metrics import roc_auc_score, precision_recall_fscore_support
from sklearn.preprocessing import OneHotEncoder

# Generate a small test dataframe with class A, B, and C probabilities
data = {
    "true_class": ["A", "A", "B", "B", "A", "A", "B", "B", "A", "C"],
    "pred_class": ["C", "C", "C", "C", "C", "C", "C", "C", "C", "C"],
    "A_proba": [0.3, 0.3, 0.2, 0.2, 0.3, 0.3, 0.2, 0.2, 0.3, 0.2],
    "B_proba": [0.1, 0.2, 0.3, 0.3, 0.2, 0.2, 0.3, 0.3, 0.2, 0.2],
    "C_proba": [0.6, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.6],
}

# Create the DataFrame
df = pd.DataFrame(data)

# Extract class labels and probabilities
class_labels = ["A", "B", "C"]
y_true = df["true_class"]

# One-hot encode the true labels
encoder = OneHotEncoder(categories=[class_labels], sparse_output=False)
y_true_encoded = encoder.fit_transform(df[["true_class"]])

# Extract predicted probabilities for each class
y_score = df[["A_proba", "B_proba", "C_proba"]]

# Calculate ROC AUC using "all-in-one" approach
roc_auc_all_in_one = roc_auc_score(y_true_encoded, y_score, multi_class="ovr", average="macro")

# Calculate ROC AUC for each class individually (per-class approach)
roc_auc_per_class = []
for i, label in enumerate(class_labels):
    y_true_class = y_true_encoded[:, i]  # True labels for this class
    y_score_class = y_score.iloc[:, i]  # Predicted probabilities for this class

    # Calculate ROC AUC for this class
    auc = roc_auc_score(y_true_class, y_score_class)
    roc_auc_per_class.append(auc)

# Sanity checks
#
# 1: Ensure that the highest probability corresponds to the predicted class (XGBoost logic)
# 2: Ensure that all the probabilities sum to 1
prediction_col = y_score.idxmax(axis=1).str.replace("_proba", "")
assert df["pred_class"].equals(prediction_col)
assert y_score.sum(axis=1).eq(1).all()

# Get the precision, recall, and f-score for each class
precision, recall, fscore, support = precision_recall_fscore_support(
    y_true, df["pred_class"], labels=class_labels, zero_division=0
)

# Format the results for each class
for i, label in enumerate(class_labels):
    print(f"\nClass: {label}")
    print(f"  Precision: {precision[i]:.2f}")
    print(f"  Recall: {recall[i]:.2f}")
    print(f"  F1-Score: {fscore[i]:.2f}")
    print(f"  ROC AUC (per class): { roc_auc_per_class[i]}")
    print(f"  Averaged ROC AUC (OVR): {roc_auc_all_in_one}")



"""An Example Model plugin component"""

import logging
from dash import dcc
import random
import plotly.graph_objects as go

# Workbench Imports
from workbench.web_interface.components.plugin_interface import PluginInterface, PluginPage, PluginInputType
from workbench.api.model import Model

# Get the Workbench logger
log = logging.getLogger("workbench")


class ModelPlugin(PluginInterface):
    """MyModelPlugin Component"""

    """Initialize this Plugin Component Class with required attributes"""
    auto_load_page = PluginPage.MODEL
    plugin_input_type = PluginInputType.MODEL

    def create_component(self, component_id: str) -> dcc.Graph:
        """Create a Model Component without any data.
        Args:
            component_id (str): The ID of the web component
        Returns:
            dcc.Graph: The EndpointTurbo Component
        """
        self.component_id = component_id
        self.container = dcc.Graph(id=component_id, figure=self.display_text("Waiting for Data..."))

        # Fill in plugin properties
        self.properties = [(self.component_id, "figure")]

        # Return the container
        return self.container

    def update_properties(self, model: Model, **kwargs) -> list:
        """Update the properties for the plugin.

        Args:
            model (Model): An instantiated Model object
            **kwargs: Additional keyword arguments (unused)

        Returns:
            list: A list of the updated property values for the plugin
        """
        log.important(f"Updating Model Plugin with Model: {model.name} and kwargs: {kwargs}")
        model_name = f"Model: {model.name}"

        # Generate random values for the pie chart
        pie_values = [random.randint(10, 30) for _ in range(4)]

        # Create a pie chart with the endpoint name as the title
        pie_figure = go.Figure(
            data=[go.Pie(labels=["A", "B", "C", "D"], values=pie_values)], layout=go.Layout(title=model_name)
        )

        # Return the updated property values for the plugin
        return [pie_figure]


if __name__ == "__main__":
    from workbench.web_interface.components.plugin_unit_test import PluginUnitTest

    # Run the Unit Test on the Plugin
    PluginUnitTest(ModelPlugin).run()



"""Example for Tautomerizing SMILES strings"""

import pandas as pd
from workbench.utils.chem_utils import tautomerize_smiles

pd.options.display.max_columns = None
pd.options.display.width = 1200


test_data = [
    # Salicylaldehyde undergoes keto-enol tautomerization.
    {"id": "Salicylaldehyde (Keto)", "smiles": "O=Cc1cccc(O)c1", "expected": "O=Cc1cccc(O)c1"},
    {"id": "2-Hydroxybenzaldehyde (Enol)", "smiles": "Oc1ccc(C=O)cc1", "expected": "O=Cc1ccc(O)cc1"},
    # Acetylacetone undergoes keto-enol tautomerization to favor the enol form.
    {"id": "Acetylacetone", "smiles": "CC(=O)CC(=O)C", "expected": "CC(=O)CC(C)=O"},
    # Imidazole undergoes a proton shift in the aromatic ring.
    {"id": "Imidazole", "smiles": "c1cnc[nH]1", "expected": "c1c[nH]cn1"},
    # Pyridone prefers the lactam form in RDKit's tautomer enumeration.
    {"id": "Pyridone", "smiles": "C1=CC=NC(=O)C=C1", "expected": "O=c1cccccn1"},
    # Guanidine undergoes amine-imine tautomerization.
    {"id": "Guanidine", "smiles": "C(=N)N=C(N)N", "expected": "N=C(N)N=CN"},
    # Catechol standardizes hydroxyl group placement in the aromatic system.
    {"id": "Catechol", "smiles": "c1cc(c(cc1)O)O", "expected": "Oc1ccccc1O"},
    # Formamide canonicalizes to NC=O, reflecting its stable form.
    {"id": "Formamide", "smiles": "C(=O)N", "expected": "NC=O"},
    # Urea undergoes a proton shift between nitrogen atoms.
    {"id": "Urea", "smiles": "C(=O)(N)N", "expected": "NC(N)=O"},
    # Phenol standardizes hydroxyl group placement in the aromatic system.
    {"id": "Phenol", "smiles": "c1ccc(cc1)O", "expected": "Oc1ccccc1"},
]

# Convert test data to a DataFrame
df = pd.DataFrame(test_data)

# Perform tautomerization
result_df = tautomerize_smiles(df)
print(result_df)



"""JSONToDataSource: Class to move local JSON Files into a Workbench DataSource"""

import os
import pandas as pd

# Local imports
from workbench.core.transforms.transform import Transform, TransformInput, TransformOutput
from workbench.core.transforms.pandas_transforms.pandas_to_data import PandasToData


class JSONToDataSource(Transform):
    """JSONToDataSource: Class to move local JSON Files into a Workbench DataSource

    Common Usage:
        ```python
        json_to_data = JSONToDataSource(json_file_path, data_name)
        json_to_data.set_output_tags(["abalone", "json", "whatever"])
        json_to_data.transform()
        ```
    """

    def __init__(self, json_file_path: str, data_name: str):
        """JSONToDataSource: Class to move local JSON Files into a Workbench DataSource

        Args:
            json_file_path (str): The path to the JSON file to be transformed
            data_name (str): The Name of the Workbench DataSource to be created
        """

        # Call superclass init
        super().__init__(json_file_path, data_name)

        # Set up all my instance attributes
        self.input_type = TransformInput.LOCAL_FILE
        self.output_type = TransformOutput.DATA_SOURCE

    def transform_impl(self, overwrite: bool = True):
        """Convert the local JSON file into Parquet Format in the Workbench Data Sources Bucket, and
        store the information about the data to the AWS Data Catalog workbench database
        """

        # Report the transformation initiation
        json_file = os.path.basename(self.input_name)
        self.log.info(f"Starting {json_file} -->  DataSource: {self.output_name}...")

        # Read in the Local JSON as a Pandas DataFrame
        df = pd.read_json(self.input_name, lines=True)

        # Use the Workbench Pandas to Data Source class
        pandas_to_data = PandasToData(self.output_name)
        pandas_to_data.set_input(df)
        pandas_to_data.set_output_tags(self.output_tags)
        pandas_to_data.add_output_meta(self.output_meta)
        pandas_to_data.transform()

        # Report the transformation results
        self.log.info(f"{json_file} -->  DataSource: {self.output_name} Complete!")

    def post_transform(self, **kwargs):
        """Post-Transform"""
        self.log.info("Post-Transform: S3 to DataSource...")

        # Note: We do not need to onboard because PandasToData already onboarded


if __name__ == "__main__":
    """Exercise the JSONToDataSource Class"""
    import sys
    from pathlib import Path

    # Get the path to the dataset in the repository data directory
    data_path = str(Path(sys.modules["workbench"].__file__).parent.parent.parent / "data" / "test_data.json")

    # Create my Data Loader
    my_loader = JSONToDataSource(data_path, "test_data_json")
    my_loader.set_output_tags(["test", "small"])

    # Store this data as a Workbench DataSource
    my_loader.transform()



"""TrainingView Class: A View with an additional training column that marks holdout ids"""

from typing import Union

# Workbench Imports
from workbench.api import DataSource, FeatureSet
from workbench.core.views.view import View
from workbench.core.views.create_view import CreateView
from workbench.core.views.view_utils import get_column_list


class TrainingView(CreateView):
    """TrainingView Class: A View with an additional training column that marks holdout ids

    Common Usage:
        ```python
        # Create a default TrainingView
        fs = FeatureSet("test_features")
        training_view = TrainingView.create(fs)
        df = training_view.pull_dataframe()

        # Create a TrainingView with a specific set of columns
        training_view = TrainingView.create(fs, column_list=["my_col1", "my_col2"])

        # Query the view
        df = training_view.query(f"SELECT * FROM {training_view.table} where training = TRUE")
        ```
    """

    @classmethod
    def create(
        cls,
        feature_set: FeatureSet,
        source_table: str = None,
        id_column: str = None,
        holdout_ids: Union[list[str], list[int], None] = None,
    ) -> Union[View, None]:
        """Factory method to create and return a TrainingView instance.

        Args:
            feature_set (FeatureSet): A FeatureSet object
            source_table (str, optional): The table/view to create the view from. Defaults to None.
            id_column (str, optional): The name of the id column. Defaults to None.
            holdout_ids (Union[list[str], list[int], None], optional): A list of holdout ids. Defaults to None.

        Returns:
            Union[View, None]: The created View object (or None if failed to create the view)
        """
        # Instantiate the TrainingView with "training" as the view name
        instance = cls("training", feature_set, source_table)

        # Drop any columns generated from AWS
        aws_cols = ["write_time", "api_invocation_time", "is_deleted", "event_time"]
        source_table_columns = get_column_list(instance.data_source, instance.source_table)
        column_list = [col for col in source_table_columns if col not in aws_cols]

        # Sanity check on the id column
        if not id_column:
            instance.log.important("No id column specified, we'll try the auto_id_column ..")
            if not instance.auto_id_column:
                instance.log.error("No id column specified and no auto_id_column found, aborting ..")
                return None
            else:
                if instance.auto_id_column not in column_list:
                    instance.log.error(
                        f"Auto id column {instance.auto_id_column} not found in column list, aborting .."
                    )
                    return None
                else:
                    id_column = instance.auto_id_column

        # If we don't have holdout ids, create a default training view
        if not holdout_ids:
            instance._default_training_view(instance.data_source, id_column)
            return View(instance.data_source, instance.view_name, auto_create_view=False)

        # Format the list of holdout ids for SQL IN clause
        if holdout_ids and all(isinstance(id, str) for id in holdout_ids):
            formatted_holdout_ids = ", ".join(f"'{id}'" for id in holdout_ids)
        else:
            formatted_holdout_ids = ", ".join(map(str, holdout_ids))

        # Enclose each column name in double quotes
        sql_columns = ", ".join([f'"{column}"' for column in column_list])

        # Construct the CREATE VIEW query
        create_view_query = f"""
        CREATE OR REPLACE VIEW {instance.table} AS
        SELECT {sql_columns}, CASE
            WHEN {id_column} IN ({formatted_holdout_ids}) THEN False
            ELSE True
        END AS training
        FROM {instance.source_table}
        """

        # Execute the CREATE VIEW query
        instance.data_source.execute_statement(create_view_query)

        # Return the View
        return View(instance.data_source, instance.view_name, auto_create_view=False)

    # This is an internal method that's used to create a default training view
    def _default_training_view(self, data_source: DataSource, id_column: str):
        """Create a default view in Athena that assigns roughly 80% of the data to training

        Args:
            data_source (DataSource): The Workbench DataSource object
            id_column (str): The name of the id column
        """
        self.log.important(f"Creating default Training View {self.table}...")

        # Drop any columns generated from AWS
        aws_cols = ["write_time", "api_invocation_time", "is_deleted", "event_time"]
        column_list = [col for col in data_source.columns if col not in aws_cols]

        # Enclose each column name in double quotes
        sql_columns = ", ".join([f'"{column}"' for column in column_list])

        # Construct the CREATE VIEW query with a simple modulo operation for the 80/20 split
        create_view_query = f"""
        CREATE OR REPLACE VIEW "{self.table}" AS
        SELECT {sql_columns}, CASE
            WHEN MOD(ROW_NUMBER() OVER (ORDER BY {id_column}), 10) < 8 THEN True  -- Assign 80% to training
            ELSE False  -- Assign roughly 20% to validation/test
        END AS training
        FROM {self.base_table_name}
        """

        # Execute the CREATE VIEW query
        data_source.execute_statement(create_view_query)


if __name__ == "__main__":
    """Exercise the Training View functionality"""
    from workbench.api import FeatureSet

    # Get the FeatureSet
    fs = FeatureSet("test_features")

    # Delete the existing training view
    training_view = TrainingView.create(fs)
    training_view.delete()

    # Create a default TrainingView
    training_view = TrainingView.create(fs)
    print(training_view)

    # Pull the training data
    df = training_view.pull_dataframe()
    print(df.head())
    print(df["training"].value_counts())

    # Create a TrainingView with holdout ids
    my_holdout_ids = list(range(10))
    training_view = TrainingView.create(fs, id_column="id", holdout_ids=my_holdout_ids)

    # Pull the training data
    df = training_view.pull_dataframe()
    print(df.head())
    print(df["training"].value_counts())
